import geopandas as gpd
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from igraph import Graph
from pandas import DataFrame
from shapely import shortest_line


def _od_cost_matrix(
    graph: Graph,
    origins: GeoDataFrame,
    destinations: GeoDataFrame,
    weight: str,
    *,
    lines: bool = False,
    rowwise: bool = False,
) -> DataFrame | GeoDataFrame:
    distances: list[list[str]] = graph.distances(
        weights="weight",
        source=origins["temp_idx"],
        target=destinations["temp_idx"],
    )

    ori_idx, des_idx, costs = [], [], []
    for i, f_idx in enumerate(origins["temp_idx"]):
        for j, t_idx in enumerate(destinations["temp_idx"]):
            ori_idx.append(f_idx)
            des_idx.append(t_idx)
            costs.append(distances[i][j])

    results = (
        pd.DataFrame(data={"origin": ori_idx, "destination": des_idx, weight: costs})
        .replace([np.inf, -np.inf], np.nan)
        .reset_index(drop=True)
    )

    # calculating all-to-all distances is much faster than looping rowwise,
    # so filtering to rowwise afterwards instead
    if rowwise:
        rowwise_df = DataFrame(
            {
                "origin": origins["temp_idx"].reset_index(drop=True),
                "destination": destinations["temp_idx"].reset_index(drop=True),
            }
        )
        results = rowwise_df.merge(results, on=["origin", "destination"], how="left")

    wkt_dict_origin = {
        idx: geom.wkt
        for idx, geom in zip(origins["temp_idx"], origins.geometry, strict=True)
    }
    wkt_dict_destination = {
        idx: geom.wkt
        for idx, geom in zip(
            destinations["temp_idx"], destinations.geometry, strict=True
        )
    }
    results["wkt_ori"] = results["origin"].map(wkt_dict_origin)
    results["wkt_des"] = results["destination"].map(wkt_dict_destination)

    results[weight] = np.where(results.wkt_ori == results.wkt_des, 0, results[weight])

    # straight lines between origin and destination
    if lines:
        origin = gpd.GeoSeries.from_wkt(results["wkt_ori"], crs=25833)
        destination = gpd.GeoSeries.from_wkt(results["wkt_des"], crs=25833)
        results["geometry"] = shortest_line(origin, destination)
        results = gpd.GeoDataFrame(results, geometry="geometry", crs=25833)

    results = results.drop(["wkt_ori", "wkt_des"], axis=1, errors="ignore")

    return results.reset_index(drop=True)



"""Overlay function that avoids a nasty GEOSException from geopandas.overlay.

This module includes the function 'clean_shapely_overlay', which bypasses a
GEOSException from the regular geopandas.overlay. The function is a generalized
version of the solution from GH 2792. The module also includes the 'overlay' function,
which first tries a geopandas.overlay, and if it raises a GEOSException, tries
'clean_shapely_overlay'.

Both 'overlay' and 'clean_shapely_overlay' also include the overlay type "update",
which can be specified in the "how" parameter, in addition to the five native geopandas
how-s.
"""

import geopandas as gpd
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from shapely import GEOSException, STRtree, difference, intersection, union_all

from .general import _push_geom_col, clean_geoms
from .geometry_types import get_geom_type, is_single_geom_type, to_single_geom_type


def overlay(
    df1: GeoDataFrame,
    df2: GeoDataFrame,
    how: str = "intersection",
    keep_geom_type: bool = True,
    geom_type: str | tuple[str, str] | None = None,
    **kwargs,
) -> GeoDataFrame:
    """Overlay that bypasses a GEOSException raised by geopandas.overlay.

    First fixes geometries and tries a regular geopandas.overlay. If it is
    unsuccessful, tries a solution from GH 2792. This solution bypasses a
    GEOSExceptions raised in the regular geopandas.overlay.

    Like in geopandas, the default is to keep only the geometry type of df1. If
    either 'df1' or 'df2' has mixed geometries, it will raise an
    Exception if not 'geom_type' is specified. 'geom_type' can be a string or a tuple
    of two strings for the geometry type of the left and right GeoDataFrame
    respectfully.

    Args:
        df1: GeoDataFrame
        df2: GeoDataFrame
        how: Method of spatial overlay. Includes the 'update' method, plus the five
            native geopandas methods 'intersection', 'union', 'identity',
            'symmetric_difference' and 'difference'.
        keep_geom_type: If True (default), return only geometries of the same
            geometry type as df1 has, if False, return all resulting geometries.
        geom_type: optionally specify what geometry type to keep before the overlay,
            if there may be mixed geometry types. Either a string with one geom_type
            or a tuple/list with geom_type for df1 and df2 respectfully.
        **kwargs: Additional keyword arguments passed to geopandas.overlay

    Returns:
        GeoDataFrame with overlayed and fixed geometries and columns from both
        GeoDataFrames.

    Raises:
        ValueError: If the geometries have mixed geometry types and 'geom_type' is not
            specified.
    """
    # Allowed operations (includes 'update')
    allowed_hows = [
        "intersection",
        "union",
        "identity",
        "symmetric_difference",
        "difference",  # aka erase
        "update",  # not in geopandas
    ]
    # Error Messages
    if how not in allowed_hows:
        raise ValueError(
            f"'how' was {how!r} but is expected to be in {', '.join(allowed_hows)}"
        )

    if df1.crs != df2.crs:
        raise ValueError(f"crs mismatch. Got {df1.crs} and {df2.crs}")

    df1 = clean_geoms(df1)
    df2 = clean_geoms(df2)

    geom_type_left, geom_type_right = _get_geom_type_left_right(geom_type)

    if geom_type_left:
        df1 = to_single_geom_type(df1, geom_type=geom_type_left)
    if geom_type_right:
        df2 = to_single_geom_type(df2, geom_type=geom_type_right)

    if not is_single_geom_type(df1):
        raise ValueError(
            "mixed geometry types in 'df1'. Specify 'geom_type' as "
            "'polygon', 'line' or 'point'."
        )
    if not is_single_geom_type(df2):
        raise ValueError(
            "mixed geometry types in 'df2'. Specify 'geom_type' as "
            "'polygon', 'line' or 'point'."
        )

    if keep_geom_type and not geom_type_left:
        geom_type_left = get_geom_type(df1)

    df1 = df1.loc[:, ~df1.columns.str.contains("index|level_")]
    df2 = df2.loc[:, ~df2.columns.str.contains("index|level_")]

    # determine what function to use
    if how == "update":
        overlayfunc = overlay_update
    else:
        overlayfunc = gpd.overlay
        kwargs = kwargs | {"how": how}

    try:
        overlayed = overlayfunc(df1, df2, **kwargs)
    except GEOSException as e:
        if how == "update":
            raise e
        overlayed = clean_shapely_overlay(df1, df2, how=how)

    overlayed = clean_geoms(overlayed)

    if keep_geom_type:
        overlayed = to_single_geom_type(overlayed, geom_type_left)

    return overlayed.loc[:, ~overlayed.columns.str.contains("index|level_")]


def overlay_update(
    df1: GeoDataFrame,
    df2: GeoDataFrame,
    keep_geom_type: bool = True,
    geom_type: str | tuple[str, str] | None = None,
    **kwargs,
) -> GeoDataFrame:
    """Updates df1 by putting df2 on top.

    First runs a difference overlay of df1 and df2 to erase the parts of df1 that is in
    df2. Then concatinates the overlayed geometries with df2.

    Args:
        df1: GeoDataFrame
        df2: GeoDataFrame
        keep_geom_type: If True (default), return only geometries of the same
            geometry type as df1 has, if False, return all resulting geometries.
        geom_type: optionally specify what geometry type to keep before the overlay,
            if there may be mixed geometry types. Either a string with one geom_type
            or a tuple/list with geom_type for df1 and df2 respectfully.
        **kwargs: Additional keyword arguments passed to geopandas.overlay

    Returns:
        GeoDataFrame with overlayed geometries and columns from both GeoDataFrames.

    """
    try:
        overlayed = df1.overlay(df2, how="difference", **kwargs)
    except GEOSException:
        overlayed = clean_shapely_overlay(
            df1,
            df2,
            how="difference",
            keep_geom_type=keep_geom_type,
            geom_type=geom_type,
        )
    overlayed = overlayed.loc[:, ~overlayed.columns.str.contains("index|level_")]
    return pd.concat([overlayed, df2], ignore_index=True)


def clean_shapely_overlay(
    df1: GeoDataFrame,
    df2: GeoDataFrame,
    how: str = "intersection",
    keep_geom_type: bool = True,
    geom_type: str | tuple[str, str] | None = None,
) -> GeoDataFrame:
    """Fixes and explodes geometries before doing a shapely overlay, then cleans up.

    Fixes geometries, then does a shapely overlay operation that dodges a GEOSException
    raised in the regular geopandas.overlay. The function is a generalised version of a
    solution from GH 2792.

    Args:
        df1: GeoDataFrame
        df2: GeoDataFrame
        how: Method of spatial overlay. Includes the 'update' method, plus the five
            native geopandas methods 'intersection', 'union', 'identity',
            'symmetric_difference' and 'difference'.
        keep_geom_type: If True (default), return only geometries of the same
            geometry type as df1 has, if False, return all resulting geometries.
        geom_type: optionally specify what geometry type to keep before the overlay,
            if there may be mixed geometry types. Either a string with one geom_type
            or a tuple/list with geom_type for df1 and df2 respectfully.

    Returns:
        GeoDataFrame with overlayed and fixed geometries and columns from both
        GeoDataFrames.

    Raises:
        ValueError: If 'how' is not one of 'intersection', 'union', 'identity',
            'symmetric_difference', 'difference' or 'update'.
    """
    # Allowed operations
    allowed_hows = [
        "intersection",
        "union",
        "identity",
        "symmetric_difference",
        "difference",  # aka erase
    ]
    # Error messages
    if how not in allowed_hows:
        raise ValueError(
            f"`how` was {how!r} but is expected to be in {', '.join(allowed_hows)}"
        )

    geom_type_left, geom_type_right = _get_geom_type_left_right(geom_type)

    if keep_geom_type and not geom_type_left:
        geom_type_left = get_geom_type(df1)

    df1 = clean_geoms(df1)
    if geom_type_left:
        df1 = to_single_geom_type(df1, geom_type_left)
    df2 = clean_geoms(df2)
    if geom_type_right:
        df2 = to_single_geom_type(df2, geom_type_right)

    df1 = df1.explode(ignore_index=True)
    df2 = df2.explode(ignore_index=True)

    overlayed = _shapely_overlay(df1, df2, how=how).pipe(clean_geoms)
    if geom_type_left:
        overlayed = to_single_geom_type(overlayed, geom_type_left)

    return overlayed.reset_index(drop=True)


def _get_geom_type_left_right(
    geom_type: str | tuple | list | None,
) -> tuple[str | None, str | None]:
    if isinstance(geom_type, (tuple, list)):
        if len(geom_type) == 1:
            return geom_type[0], geom_type[0]
        elif len(geom_type) == 2:
            return geom_type
        else:
            raise ValueError(
                "'geom_type' should be one or two strings for the left and right gdf"
            )
    elif isinstance(geom_type, str):
        return geom_type, geom_type
    elif geom_type is None:
        return None, None
    else:
        raise ValueError(
            "'geom_type' should be one or two strings for the left and right gdf"
        )


def _shapely_overlay(df1: GeoDataFrame, df2: GeoDataFrame, how: str) -> GeoDataFrame:
    tree = STRtree(df2.geometry.values)
    left, right = tree.query(df1.geometry.values, predicate="intersects")

    pairs = _get_intersects_pairs(df1, df2, left, right)

    if how == "intersection":
        return _intersection(pairs)

    if how == "difference":
        return _difference(pairs, df1, left)

    if how == "symmetric_difference":
        return _symmetric_difference(pairs, df1, df2, left, right)

    if how == "identity":
        return _identity(pairs, df1, left)

    if how == "union":
        return _union(pairs, df1, df2, left, right)


def _intersection(pairs: GeoDataFrame) -> GeoDataFrame:
    intersections = pairs.copy()
    intersections["geometry"] = intersection(
        intersections.geometry.values, intersections.geom_right.values
    )
    intersections = intersections.drop(columns=["index_right", "geom_right"])
    return intersections


def _union(pairs, df1, df2, left, right):
    merged = []
    if len(left):
        intersections = _intersection(pairs)
        merged.append(intersections)
    symmdiff = _symmetric_difference(pairs, df1, df2, left, right)
    merged.append(symmdiff)
    crs = merged[0].crs
    merged = [gdf.to_crs(crs) for gdf in merged]
    return pd.concat(merged, ignore_index=True).pipe(_push_geom_col)


def _identity(pairs, df1, left):
    merged = []
    if len(left):
        intersections = _intersection(pairs)
        merged.append(intersections)
    diff = _difference(pairs, df1, left)
    merged.append(diff)
    crs = merged[0].crs
    merged = [gdf.to_crs(crs) for gdf in merged]
    return pd.concat(merged, ignore_index=True).pipe(_push_geom_col)


def _symmetric_difference(pairs, df1, df2, left, right):
    merged = []
    difference_left = _difference(pairs, df1, left)
    merged.append(difference_left)
    if len(left):
        clip_right = _shapely_diffclip_right(pairs, df1, df2)
        merged.append(clip_right)
    diff_right = _add_from_right(df1, df2, right)
    merged.append(diff_right)
    crs = merged[0].crs
    merged = [gdf.to_crs(crs) for gdf in merged]
    return pd.concat(merged, ignore_index=True).pipe(_push_geom_col)


def _difference(pairs, df1, left):
    merged = []
    if len(left):
        clip_left = _shapely_diffclip_left(pairs, df1)
        merged.append(clip_left)
    diff_left = _add_from_left(df1, left)
    merged.append(diff_left)
    crs = merged[0].crs
    merged = [gdf.to_crs(crs) for gdf in merged]
    return pd.concat(merged, ignore_index=True).pipe(_push_geom_col)


def _get_intersects_pairs(
    df1: GeoDataFrame, df2: GeoDataFrame, left: np.ndarray, right: np.ndarray
) -> GeoDataFrame:
    return pd.concat(
        [
            df1.take(left),
            (pd.DataFrame({"index_right": right}, index=df1.index.values.take(left))),
        ],
        axis=1,
    ).join(
        df2.rename(columns={"geometry": "geom_right"}),
        on="index_right",
        rsuffix="_2",
    )


def _add_from_left(df1, left):
    return df1.take(np.setdiff1d(np.arange(len(df1)), left))


def _add_from_right(
    df1: GeoDataFrame, df2: GeoDataFrame, right: np.ndarray
) -> GeoDataFrame:
    return df2.take(np.setdiff1d(np.arange(len(df2)), right)).rename(
        columns={
            c: f"{c}_2" if c in df1.columns and c != "geometry" else c
            for c in df2.columns
        }
    )


def _shapely_diffclip_left(pairs, df1):
    """Aggregate areas in right by unique values of left, then use those to clip
    areas out of left"""
    clip_left = gpd.GeoDataFrame(
        pairs.groupby(level=0).agg(
            {
                "geom_right": lambda g: union_all(g) if len(g) > 1 else g,
                **{
                    c: "first"
                    for c in df1.columns
                    if c not in ["index_right", "geom_right"]
                },
            }
        ),
        geometry="geometry",
        crs=df1.crs,
    )
    clip_left["geometry"] = difference(
        clip_left.geometry.values, clip_left.geom_right.values
    )
    clip_left = clip_left.drop(columns=["geom_right"])

    return clip_left


def _shapely_diffclip_right(pairs, df1, df2):
    clip_right = (
        gpd.GeoDataFrame(
            pairs.rename(columns={"geometry": "geom_left", "geom_right": "geometry"})
            .groupby(by="index_right")
            .agg(
                {
                    "geom_left": lambda g: union_all(g) if len(g) > 1 else g,
                    "geometry": "first",
                }
            ),
            geometry="geometry",
            crs=df2.crs,
        )
        .join(df2.drop(columns=["geometry"]))
        .rename(
            columns={
                c: f"{c}_2" if c in df1.columns and c != "geometry" else c
                for c in df2.columns
            }
        )
    )
    clip_right["geometry"] = difference(
        clip_right.geometry.values, clip_right.geom_left.values
    )
    clip_right = clip_right.drop(columns=["geom_left"])
    return clip_right



"""DirectedNetwork class for directed network analysis."""

import warnings

import pandas as pd
from geopandas import GeoDataFrame
from shapely.constructive import reverse

from ..helpers import return_two_vals, unit_is_meters
from .network import Network


class DirectedNetwork(Network):
    """Class for preparing line data for directed network analysis.

    Can be used as the 'network' parameter in the NetworkAnalysis class for directed
    network analysis.

    The DirectedNetwork class has all the methods of the base Network class,
    plus additional methods for making the network directed. If a DirectedNetwork
    instance is used in NetworkAnalysis, the network graph will be directed,
    meaning the lines can only be traversed in one direction. The network should
    be make properly directed with the 'make_directed_network' method, or
    'make_directed_network_norway' for Norwegian road data.

    Args:
        gdf: a GeoDataFrame of line geometries.
        **kwargs: keyword arguments taken by the base class Network.

    Attributes:
        gdf: the GeoDataFrame of lines with source and target ids.
        nodes: GeoDataFrame of points with unique ids. The ids are dynamic, and changes
            every time the lines change. The nodes cannot be changed directly.

    Examples
    --------
    Read example data.

    >>> from sgis import read_parquet_url, DirectedNetwork
    >>> roads = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_eidskog_2022.parquet")

    Create a DirectedNetwork instance. Notice that percent_bidirectional is 0.

    >>> nw = DirectedNetwork(roads).remove_isolated()
    >>> nw
    DirectedNetwork(3407 km, percent_bidirectional=0)

    Make the network bidirectional by specifying the direction column and its values.

    >>> nw = nw.make_directed_network(
    ...     direction_col="oneway",
    ...     direction_vals_bft=("B", "FT", "TF"),
    ...     minute_cols=("drivetime_fw", "drivetime_bw"),
    ... )
    >>> nw
    DirectedNetwork(6364 km, percent_bidirectional=87)

    Custom method for Norwegian road data (https://kartkatalog.geonorge.no/metadata/nvdb-ruteplan-nettverksdatasett/8d0f9066-34f9-4423-be12-8e8523089313),
    which does the same as the example above.

    >>> nw = (
    ...     DirectedNetwork(roads)
    ...     .remove_isolated()
    ...     .make_directed_network_norway()
    ...     )
    >>> nw
    DirectedNetwork(6364 km, percent_bidirectional=87)
    """

    def __init__(
        self,
        gdf: GeoDataFrame,
        **kwargs,
    ):
        """Initialiser passed to the Network base class."""
        super().__init__(gdf, **kwargs)

        # for the DirectedNetwork class, the graph will be directed in network analysis
        self._as_directed = True

    def make_directed_network(
        self,
        direction_col: str,
        direction_vals_bft: tuple[str, str, str],
        minute_cols: tuple[str, str] | str | None = None,
        speed_col: str | None = None,
        flat_speed: int | None = None,
        reverse_tofrom: bool = True,
    ):
        """Flips the line geometries of roads going backwards and in both directions.

        Args:
            direction_col: name of column specifying the direction of the line
                geometry.
            direction_vals_bft: tuple or list with the values of the direction
                column. Must be in the order 'both directions', 'from', 'to'.
                E.g. ('B', 'F', 'T').
            minute_cols (optional): column or columns containing the number of minutes
                it takes to traverse the line. If one column name is given, this will
                be used for both directions. If tuple/list with two column names,
                the first column will be used as the minute column for the forward
                direction, and the second column for roads going backwards.
            speed_col (optional): name of column with the road speed limit.
            flat_speed (optional): Speed in kilometers per hour to use as the speed for
                all roads.
            reverse_tofrom: If the geometries of the lines going backwards
                (i.e. has the last value in 'direction_vals_bft'). Defaults to True.

        Returns:
            The Network class, with the network attribute updated with flipped
                geometries for lines going backwards and both directions.
            Adds the column 'minutes' if either 'speed_col', 'minute_col' or
                'flat_speed' is specified.

        Raises:
            ValueError: If 'flat_speed' or 'speed_col' is specified and the unit of the
                coordinate reference system is not 'metre'
        """
        self._validate_minute_args(minute_cols, speed_col, flat_speed)

        self._validate_direction_args(direction_col, direction_vals_bft)

        if (flat_speed or speed_col) and not unit_is_meters(self.gdf):
            raise ValueError(
                "The crs must have 'metre' as units when calculating minutes."
                "Change crs or calculate minutes manually."
            )

        gdf = self.gdf
        b, f, t = direction_vals_bft

        if minute_cols:
            gdf = gdf.drop("minutes", axis=1, errors="ignore")

        # select the directional and bidirectional rows.
        ft = gdf.loc[gdf[direction_col] == f]
        tf = gdf.loc[gdf[direction_col] == t]
        both_ways = gdf.loc[gdf[direction_col] == b]
        both_ways2 = both_ways.copy()

        if minute_cols:
            try:
                min_f, min_t = return_two_vals(minute_cols)
            except ValueError as e:
                raise ValueError(
                    "'minute_cols' should be column name (string) or tuple/list with "
                    "values of directions forwards and backwards, in that order."
                ) from e

            # rename the two minute cols
            both_ways = both_ways.rename(columns={min_f: "minutes"}, errors="raise")
            both_ways2 = both_ways2.rename(columns={min_t: "minutes"}, errors="raise")

            ft = ft.rename(columns={min_f: "minutes"}, errors="raise")
            tf = tf.rename(columns={min_t: "minutes"}, errors="raise")

            for gdf in [ft, tf, both_ways, both_ways2]:
                if all(gdf["minutes"].fillna(0) <= 0):
                    raise ValueError("All values in minute col is NaN or less than 0.")

        both_ways2.geometry = reverse(both_ways2.geometry)

        if reverse_tofrom:
            tf.geometry = reverse(tf.geometry)

        self.gdf = pd.concat([both_ways, both_ways2, ft, tf], ignore_index=True)

        if speed_col:
            self._get_speed_from_col(speed_col)

        if flat_speed:
            self.gdf["minutes"] = self.gdf.length / flat_speed * 16.6666666667

        if "minutes" in self.gdf.columns:
            self.gdf = self.gdf.loc[self.gdf["minutes"] >= 0]

        self._make_node_ids()

        self._percent_bidirectional = self._check_percent_bidirectional()

        return self

    def make_directed_network_norway(
        self,
        direction_col: str = "oneway",
        direction_vals_bft: tuple[str, str, str] = ("B", "FT", "TF"),
        minute_cols: tuple[str, str] = ("drivetime_fw", "drivetime_bw"),
    ):
        """Runs method make_directed_network for Norwegian road data.

        Runs the method make_directed_network with default arguments to fit
        Norwegian road data:
        https://kartkatalog.geonorge.no/metadata/nvdb-ruteplan-nettverksdatasett/8d0f9066-34f9-4423-be12-8e8523089313
        """
        return self.make_directed_network(
            direction_col=direction_col,
            direction_vals_bft=direction_vals_bft,
            minute_cols=minute_cols,
        )

    def _warn_if_undirected(self):
        """Road data often have to be duplicated and flipped to make it directed."""
        if self.percent_bidirectional > 5:
            return

        mess = (
            "Your network is likely not directed. "
            f"Only {self.percent_bidirectional:.1f} percent of the lines go both ways."
        )
        if "oneway" in [col.lower() for col in self.gdf.columns]:
            mess = mess + (
                " Try setting direction_col='oneway' in the 'make_directed_network' "
                "method"
            )
        else:
            mess = mess + "Try running 'make_directed_network'"

        warnings.warn(mess, stacklevel=2)

    @staticmethod
    def _validate_minute_args(minute_cols, speed_col, flat_speed):
        if not minute_cols and not speed_col and not flat_speed:
            warnings.warn(
                "Minute column will not be calculated when both 'minute_cols', "
                "'speed_col' and 'flat_speed' is None",
                stacklevel=2,
            )

        if sum([bool(minute_cols), bool(speed_col), bool(flat_speed)]) > 1:
            raise ValueError(
                "Can only calculate minutes from either 'speed_col', "
                "'minute_cols' or 'flat_speed'."
            )

    def _validate_direction_args(self, direction_col, direction_vals_bft):
        if len(direction_vals_bft) != 3:
            raise ValueError(
                "'direction_vals_bft' should be tuple/list with values of directions "
                "both, from and to. E.g. ('B', 'F', 'T')"
            )

        gdf = self.gdf
        b, f, t = direction_vals_bft

        if not all(gdf.loc[gdf[direction_col].isin(direction_vals_bft)]):
            raise ValueError(
                f"direction_col '{direction_col}' should have only three unique",
                f"values. unique values in {direction_col}:",
                ", ".join(gdf[direction_col].fillna("nan").unique()),
                f"Got {direction_vals_bft}.",
            )

        if "b" in t.lower() and "t" in b.lower() and "f" in f.lower():
            warnings.warn(
                f"The 'direction_vals_bft' should be in the order 'both ways', "
                f"'from', 'to'. Got {b, f, t}. Is this correct?",
                stacklevel=2,
            )

    def _get_speed_from_col(self, speed_col):
        if (
            len(self.gdf.loc[(self.gdf[speed_col].isna()) | (self.gdf[speed_col] == 0)])
            > len(self.gdf) * 0.05
        ):
            raise ValueError(
                f"speed_col {speed_col!r} has a lot of missing or 0 values. Fill these "
                "with appropriate values"
            )
        self.gdf["minutes"] = (
            self.gdf.length / self.gdf[speed_col].astype(float) * 16.6666666667
        )

    def _make_directed_network_osm(
        self,
        direction_col: str = "oneway",
        direction_vals_bft: tuple[str, str, str] = ("B", "F", "T"),
        speed_col: str = "maxspeed",
    ):
        """Currently not working."""
        return self.make_directed_network(
            direction_col=direction_col,
            direction_vals_bft=direction_vals_bft,
            speed_col=speed_col,
        )

    def __repr__(self) -> str:
        """The print representation."""
        cl = self.__class__.__name__
        km = int(sum(self.gdf.length) / 1000)
        return f"{cl}({km} km, percent_bidirectional={self.percent_bidirectional})"



"""Functions for polygon geometries."""

import geopandas as gpd
import numpy as np
from geopandas import GeoDataFrame, GeoSeries
from shapely import (
    Geometry,
    area,
    get_exterior_ring,
    get_interior_ring,
    get_num_interior_rings,
    get_parts,
    polygons,
)
from shapely.ops import unary_union


def close_small_holes(
    gdf: GeoDataFrame | GeoSeries | Geometry,
    max_area: int | float,
    *,
    copy: bool = True,
) -> GeoDataFrame | GeoSeries | Geometry:
    """Closes holes in polygons if the area is less than the given maximum.

    It takes a GeoDataFrame, GeoSeries or shapely geometry of polygons object and
    fills the holes that are smaller than the specified area given in units of
    either square meters ('max_m2') or square kilometers ('max_km2').

    Args:
        gdf: GeoDataFrame, GeoSeries or shapely Geometry.
        max_area: The maximum area in the unit of the GeoDataFrame's crs.
        copy: if True (default), the input GeoDataFrame or GeoSeries is copied.
            Defaults to True.

    Returns:
        A GeoDataFrame, GeoSeries or shapely Geometry with closed holes in the geometry
        column.

    Raises:
        ValueError: If the coordinate reference system of the GeoDataFrame is not in
            meter units.
        ValueError: If both 'max_m2' and 'max_km2' is given.

    Examples
    --------

    Let's create a circle with a hole in it.

    >>> from sgis import close_small_holes, buff, to_gdf
    >>> point = to_gdf([260000, 6650000], crs=25833)
    >>> point
                            geometry
    0  POINT (260000.000 6650000.000)
    >>> circle = buff(point, 1000)
    >>> small_circle = buff(point, 500)
    >>> circle_with_hole = circle.overlay(small_circle, how="difference")
    >>> circle_with_hole.area
    0    2.355807e+06
    dtype: float64

    Close holes smaller than 1 square kilometer (1 million square meters).

    >>> holes_closed = close_small_holes(circle_with_hole, max_area=1_000_000)
    >>> holes_closed.area
    0    3.141076e+06
    dtype: float64

    The hole will not be closed if it is larger.

    >>> holes_closed = close_small_holes(circle_with_hole, max_area=1_000)
    >>> holes_closed.area
    0    2.355807e+06
    dtype: float64
    """
    if copy:
        gdf = gdf.copy()

    if isinstance(gdf, GeoDataFrame):
        gdf["geometry"] = gdf.geometry.map(
            lambda x: _close_small_holes_poly(x, max_area)
        )

    elif isinstance(gdf, gpd.GeoSeries):
        gdf = gdf.map(lambda x: _close_small_holes_poly(x, max_area))
        gdf = gpd.GeoSeries(gdf)

    else:
        gdf = _close_small_holes_poly(gdf, max_area)

    return gdf


def close_all_holes(
    gdf: GeoDataFrame | GeoSeries | Geometry,
    *,
    copy: bool = True,
) -> GeoDataFrame | GeoSeries | Geometry:
    """Closes all holes in polygons.

    It takes a GeoDataFrame, GeoSeries or shapely geometry of polygons object and
    returns the outer circle.

    Args:
        gdf: GeoDataFrame, GeoSeries or shapely Geometry.
        copy: if True (default), the input GeoDataFrame or GeoSeries is copied.
            Defaults to True.

    Returns:
        A GeoDataFrame, GeoSeries or shapely Geometry with closed holes in the geometry
        column.

    Examples
    --------
    Let's create a circle with a hole in it.

    >>> from sgis import close_all_holes, buff, to_gdf
    >>> point = to_gdf([260000, 6650000], crs=25833)
    >>> point
                            geometry
    0  POINT (260000.000 6650000.000)
    >>> circle = buff(point, 1000)
    >>> small_circle = buff(point, 500)
    >>> circle_with_hole = circle.overlay(small_circle, how="difference")
    >>> circle_with_hole.area
    0    2.355807e+06
    dtype: float64

    Close the hole.

    >>> holes_closed = close_all_holes(circle_with_hole)
    >>> holes_closed.area
    0    3.141076e+06
    dtype: float64
    """
    if copy:
        gdf = gdf.copy()

    def close_all_holes_func(poly):
        return unary_union(polygons(get_exterior_ring(get_parts(poly))))

    close_all_holes_func = np.vectorize(close_all_holes_func)

    if isinstance(gdf, GeoDataFrame):
        gdf["geometry"] = close_all_holes_func(gdf.geometry)

        return gdf

    elif isinstance(gdf, gpd.GeoSeries):
        return close_all_holes_func(gdf)

    else:
        return polygons(get_exterior_ring(get_parts(gdf)))


def _close_small_holes_poly(poly, max_area):
    """Closes cmall holes within one shapely geometry of polygons."""

    # start with a list containing the polygon,
    # then append all holes smaller than 'max_km2' to the list.
    holes_closed = [poly]
    singlepart = get_parts(poly)
    for part in singlepart:
        n_interior_rings = get_num_interior_rings(part)

        if not (n_interior_rings):
            continue

        for n in range(n_interior_rings):
            hole = polygons(get_interior_ring(part, n))

            if area(hole) < max_area:
                holes_closed.append(hole)

    return unary_union(holes_closed)



"""Functions that buffer, dissolve and/or explodes geometries while fixing geometries.

Functions with the purpose of making the code cleaner when buffering, dissolving,
exploding and repairing geometries. The functions are identical to doing buffer,
dissolve and explode individually, except for the following:

- Geometries are always repaired after buffer and dissolve.

- The buffer resolution defaults to 50, while geopandas' default is 16.

- The buff function returns a GeoDataFrame, the geopandas method returns a GeoSeries.

- index_parts is set to False, which will be the default value in a future version of geopandas.
"""

from geopandas import GeoDataFrame


IGNORE_INDEX_ERROR_MESSAGE = (
    "Cannot set ignore_index. Set as_index=False to reset the index and keep "
    "the 'by' columns. Or use reset_index(drop=True) to remove the 'by' "
    "columns completely"
)


def buffdissexp(
    gdf: GeoDataFrame,
    distance,
    *,
    resolution=50,
    index_parts: bool = False,
    copy: bool = True,
    **dissolve_kwargs,
) -> GeoDataFrame:
    """Buffers and dissolves overlapping geometries.

    It takes a GeoDataFrame and buffer, fixes, dissolves, fixes and explodes geometries.

    Args:
        gdf: the GeoDataFrame that will be buffered, dissolved and exploded.
        distance: the distance (meters, degrees, depending on the crs) to buffer
            the geometry by
        resolution: The number of segments used to approximate a quarter circle.
            Here defaults to 50, as opposed to the default 16 in geopandas.
        index_parts: If False (default), the index after dissolve is respected. If
            True, an integer index level is added during explode.
        copy: Whether to copy the GeoDataFrame before buffering.
        **dissolve_kwargs: additional keyword arguments passed to geopandas' dissolve.

    Returns:
        A buffered GeoDataFrame where overlapping geometries are dissolved.

    """
    if "ignore_index" in dissolve_kwargs:
        raise ValueError(IGNORE_INDEX_ERROR_MESSAGE)

    geom_col = gdf._geometry_column_name

    buffered = buff(gdf, distance, resolution=resolution, copy=copy)

    dissolved = buffered.dissolve(**dissolve_kwargs)

    dissolved[geom_col] = dissolved.make_valid()

    return dissolved.explode(index_parts=index_parts)


def buffdiss(
    gdf: GeoDataFrame,
    distance,
    resolution=50,
    copy: bool = True,
    **dissolve_kwargs,
) -> GeoDataFrame:
    """Buffers and dissolves geometries.

    It takes a GeoDataFrame and buffer, fixes, dissolves and fixes geometries.

    Args:
        gdf: the GeoDataFrame that will be
            buffered and dissolved.
        distance: the distance (meters, degrees, depending on the crs) to buffer
            the geometry by
        resolution: The number of segments used to approximate a quarter circle.
            Here defaults to 50, as opposed to the default 16 in geopandas.
        copy: Whether to copy the GeoDataFrame before buffering.
        **dissolve_kwargs: additional keyword arguments passed to geopandas' dissolve.

    Returns:
        A buffered GeoDataFrame where geometries are dissolved.

    Examples
    --------
    Create some random points.

    >>> import sgis as sg
    >>> import numpy as np
    >>> points = sg.read_parquet_url(
    ...     "https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/points_oslo.parquet"
    ... )[["geometry"]]
    >>> points["group"] = np.random.choice([*"abd"], len(points))
    >>> points["number"] = np.random.random(size=len(points))
    >>> points
                               geometry group    number
    0    POINT (263122.700 6651184.900)     a  0.878158
    1    POINT (272456.100 6653369.500)     a  0.693311
    2    POINT (270082.300 6653032.700)     b  0.323960
    3    POINT (259804.800 6650339.700)     a  0.606745
    4    POINT (272876.200 6652889.100)     a  0.194360
    ..                              ...   ...       ...
    995  POINT (266801.700 6647844.500)     a  0.814424
    996  POINT (261274.000 6653593.400)     b  0.769479
    997  POINT (263542.900 6645427.000)     a  0.925991
    998  POINT (269226.700 6650628.000)     b  0.431972
    999  POINT (264570.300 6644239.500)     d  0.555239

    Buffer by 100 meters and dissolve.

    >>> sg.buffdiss(points, 100)
                                                geometry group    number
    0  MULTIPOLYGON (((256421.833 6649878.117, 256420...     d  0.580157

    Dissolve by 'group' and get sum of columns.

    >>> sg.buffdiss(points, 100, by="group", aggfunc="sum")
                                                    geometry      number
    group
    a      MULTIPOLYGON (((258866.258 6648220.031, 258865...  167.265619
    b      MULTIPOLYGON (((258404.858 6647830.931, 258404...  171.939169
    d      MULTIPOLYGON (((258180.258 6647935.731, 258179...  156.964300

    To get the 'by' columns as columns, not index.

    >>> sg.buffdiss(points, 100, by="group", as_index=False)
      group                                           geometry    number
    0     a  MULTIPOLYGON (((258866.258 6648220.031, 258865...  0.323948
    1     b  MULTIPOLYGON (((258404.858 6647830.931, 258404...  0.687635
    2     d  MULTIPOLYGON (((258180.258 6647935.731, 258179...  0.580157
    """
    if "ignore_index" in dissolve_kwargs:
        raise ValueError(IGNORE_INDEX_ERROR_MESSAGE)

    geom_col = gdf._geometry_column_name

    buffered = buff(gdf, distance, resolution=resolution, copy=copy)

    dissolved = buffered.dissolve(**dissolve_kwargs)

    dissolved[geom_col] = dissolved.make_valid()

    return dissolved


def dissexp(
    gdf: GeoDataFrame,
    index_parts: bool = False,
    **dissolve_kwargs,
):
    """Dissolves overlapping geometries.

    It takes a GeoDataFrame and dissolves, fixes and explodes geometries.

    Args:
        gdf: the GeoDataFrame that will be buffered, dissolved and exploded.
        index_parts: If False (default), the index after dissolve is respected. If
            True, an integer index level is added during explode.
        **dissolve_kwargs: additional keyword arguments passed to geopandas' dissolve.

    Returns:
        A GeoDataFrame where overlapping geometries are dissolved.
    """
    if "ignore_index" in dissolve_kwargs:
        raise ValueError(IGNORE_INDEX_ERROR_MESSAGE)

    geom_col = gdf._geometry_column_name

    dissolved = gdf.dissolve(**dissolve_kwargs)

    dissolved[geom_col] = dissolved.make_valid()

    return dissolved.explode(index_parts=index_parts)


def buffexp(
    gdf,
    distance,
    resolution=50,
    index_parts: bool = False,
    copy: bool = True,
    **buffer_kwargs,
):
    """Buffers and explodes geometries.

    It takes a GeoDataFrame and buffer, fixes and explodes geometries.

    Args:
        gdf: the GeoDataFrame that will be buffered, dissolved and exploded.
        distance: the distance (meters, degrees, depending on the crs) to buffer
            the geometry by
        resolution: The number of segments used to approximate a quarter circle.
            Here defaults to 50, as opposed to the default 16 in geopandas.
        index_parts: If False (default), the index after dissolve is respected. If
            True, an integer index level is added during explode.
        copy: Whether to copy the GeoDataFrame before buffering.
        **buffer_kwargs: additional keyword arguments passed to geopandas' buffer.

    Returns:
        A buffered GeoDataFrame where geometries are exploded.
    """
    if "ignore_index" in buffer_kwargs:
        raise ValueError(IGNORE_INDEX_ERROR_MESSAGE)

    return buff(
        gdf, distance, resolution=resolution, copy=copy, **buffer_kwargs
    ).explode(index_parts=index_parts)


def buff(
    gdf: GeoDataFrame,
    distance: int | float,
    resolution: int = 50,
    copy: bool = True,
    **buffer_kwargs,
) -> GeoDataFrame:
    """Buffers a GeoDataFrame with high resolution and returns a new GeoDataFrame.

    Args:
        gdf: the GeoDataFrame that will be buffered, dissolved and exploded.
        distance: the distance (meters, degrees, depending on the crs) to buffer
            the geometry by
        resolution: The number of segments used to approximate a quarter circle.
            Here defaults to 50, as opposed to the default 16 in geopandas.
        copy: Whether to copy the GeoDataFrame before buffering.
        **buffer_kwargs: additional keyword arguments passed to geopandas' buffer.

    Returns:
        A buffered GeoDataFrame.
    """
    geom_col = gdf._geometry_column_name

    if copy:
        gdf = gdf.copy()

    gdf[geom_col] = gdf.buffer(
        distance, resolution=resolution, **buffer_kwargs
    ).make_valid()

    return gdf



"""Functions for line geometries.

The module includes functions for cutting and splitting lines, cutting lines into
pieces, filling holes in a network of lines, finding isolated network islands and
creating unique node ids.

The functions are also methods of the Network class, where some checks and
preperation is done before each method is run, making sure the lines are correct.
"""
import warnings

import geopandas as gpd
import networkx as nx
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame, GeoSeries
from pandas import DataFrame, Series
from shapely import force_2d, shortest_line
from shapely.geometry import LineString, Point

from .buffer_dissolve_explode import buff
from .general import _push_geom_col, coordinate_array, to_gdf
from .neighbors import get_k_nearest_neighbors, k_nearest_neighbors
from .point_operations import snap_all, snap_within_distance


def get_largest_component(lines: GeoDataFrame) -> GeoDataFrame:
    """Finds the largest network component.

    It takes a GeoDataFrame of lines and finds the lines that are
    part of the largest connected network component. These lines are given the value
    1 in the added column 'connected', while isolated network islands get the value
    0.

    Uses the connected_components function from the networkx package.

    Args:
        lines: A GeoDataFrame of lines.

    Returns:
        The GeoDataFrame with a new column "connected".

    Examples
    --------
    >>> from sgis import read_parquet_url, get_largest_component
    >>> roads = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_oslo_2022.parquet")

    >>> roads = get_largest_component(roads)
    >>> roads.connected.value_counts()
    1.0    85638
    0.0     7757
    Name: connected, dtype: int64
    """
    lines, _ = make_node_ids(lines)

    edges = [
        (str(source), str(target))
        for source, target in zip(lines["source"], lines["target"], strict=True)
    ]

    graph = nx.Graph()
    graph.add_edges_from(edges)

    largest_component = max(nx.connected_components(graph), key=len)

    largest_component_dict = {node_id: 1 for node_id in largest_component}

    lines["connected"] = lines.source.map(largest_component_dict).fillna(0)

    return lines


def get_component_size(lines: GeoDataFrame) -> GeoDataFrame:
    """Finds the size of each component in the network.

    Takes a GeoDataFrame of linea and creates the column "component_size", which
    indicates the size of the network component the line is a part of.

    Args:
        lines: a GeoDataFrame of lines.

    Returns:
        A GeoDataFrame with a new column "component_size".

    Examples
    --------
    >>> from sgis import read_parquet_url, get_component_size
    >>> roads = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_oslo_2022.parquet")

    >>> roads = get_component_size(roads)
    >>> roads.component_size.value_counts().head()
    79180    85638
    2         1601
    4          688
    6          406
    3          346
    Name: component_size, dtype: int64
    """
    lines, _ = make_node_ids(lines)

    edges = [
        (str(source), str(target))
        for source, target in zip(lines["source"], lines["target"], strict=True)
    ]

    graph = nx.Graph()
    graph.add_edges_from(edges)
    components = [list(x) for x in nx.connected_components(graph)]

    componentsdict = {
        idx: len(component) for component in components for idx in component
    }

    lines["component_size"] = lines.source.map(componentsdict)

    return lines


def split_lines_by_nearest_point(
    lines: GeoDataFrame,
    points: GeoDataFrame,
    max_distance: int | None = None,
) -> DataFrame:
    """Split lines that are closest to s point.

    Snaps points to nearest lines and splits the lines in two at the snap point.
    The splitting is done pointwise, meaning each point splits one line in two.
    The line will not be split if the point is closest to the endpoint of the line.

    This function is used in NetworkAnalysis if split_lines is set to True.

    Args:
        lines: GeoDataFrame of lines that will be split.
        points: GeoDataFrame of points to split the lines with.
        max_distance: the maximum distance between the point and the line. Points further
            away than max_distance will not split any lines. Defaults to None.

    Returns:
        A GeoDataFrame with the same columns as the input lines, but with the lines
        split at the closest point to the points.

    Raises:
        ValueError: If the crs of the input data differs.

    Examples
    --------
    >>> from sgis import read_parquet_url, split_lines_by_nearest_point
    >>> roads = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_oslo_2022.parquet")
    >>> points = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/points_oslo.parquet")
    >>> rows = len(roads)
    >>> rows
    93395

    Splitting lines for points closer than 10 meters from the lines.

    >>> roads = split_lines_by_nearest_point(roads, points, max_distance=10)
    >>> print("number of lines that were split:", len(roads) - rows)
    number of lines that were split: 380

    Splitting lines by all points.

    >>> roads = split_lines_by_nearest_point(roads, points)
    >>> print("number of lines that were split:", len(roads) - rows)
    number of lines that were split: 848

    Not all lines were split. That is because some points were closest to an endpoint
    of a line.
    """
    BUFFDIST = 0.000001

    if points.crs != lines.crs:
        raise ValueError("crs mismatch:", points.crs, "and", lines.crs)

    lines["temp_idx_"] = lines.index

    # move the points to the nearest exact point of the line
    if max_distance:
        snapped = snap_within_distance(points, lines, max_distance=max_distance)
    else:
        snapped = snap_all(points, lines)

    # find the lines that were snapped to (or are very close)
    snapped_buff = buff(snapped, BUFFDIST)
    intersect = lines.intersects(snapped_buff.unary_union)
    relevant_lines = lines.loc[intersect]
    the_other_lines = lines.loc[~intersect]

    # need consistent coordinate dimensions later
    # (doing it down here to not overwrite the original data)
    relevant_lines.geometry = force_2d(relevant_lines.geometry)
    snapped.geometry = force_2d(snapped.geometry)

    # split the lines with buffer + difference, since shaply.split usually doesn't work
    splitted = relevant_lines.overlay(snapped_buff, how="difference").explode(
        ignore_index=True
    )

    # the endpoints of the new lines are now sligtly off. To get the exact snapped
    # point coordinates, using get_k_nearest_neighbors. This will map the sligtly
    # off line endpoints with the point the line was split by.

    snapped["point_coords"] = [(geom.x, geom.y) for geom in snapped.geometry]

    # get the endpoints of the lines as columns
    splitted = make_edge_coords_cols(splitted)

    splitted_source = to_gdf(splitted["source_coords"], crs=lines.crs)
    splitted_target = to_gdf(splitted["target_coords"], crs=lines.crs)

    # find the nearest snapped point for each source and target of the lines
    snapped = snapped.set_index("point_coords")
    dists_source: DataFrame = get_k_nearest_neighbors(splitted_source, snapped, k=1)
    dists_target: DataFrame = get_k_nearest_neighbors(splitted_target, snapped, k=1)

    dists_source = dists_source.loc[dists_source.distance <= BUFFDIST * 2]
    dists_target = dists_target.loc[dists_target.distance <= BUFFDIST * 2]

    pointmapper_source: pd.Series = dists_source["neighbor_index"]
    pointmapper_target: pd.Series = dists_target["neighbor_index"]

    # now, we can finally replace the source/target coordinate with the coordinates of
    # the snapped points.

    # loop for each line where the source is the endpoint that was split
    # change the first point of the line to the point it was split by
    for idx in dists_source.index:
        line = splitted.loc[idx, "geometry"]
        coordslist = list(line.coords)
        coordslist[0] = pointmapper_source[idx]
        splitted.loc[idx, "geometry"] = LineString(coordslist)

    # same for the lines where the target was split, but change the last point of the
    # line
    for idx in dists_target.index:
        line = splitted.loc[idx, "geometry"]
        coordslist = list(line.coords)
        coordslist[-1] = pointmapper_target[idx]
        splitted.loc[idx, "geometry"] = LineString(coordslist)

    splitted["splitted"] = 1

    lines = pd.concat([the_other_lines, splitted], ignore_index=True).drop(
        ["temp_idx_", "source_coords", "target_coords"], axis=1
    )

    return lines


def cut_lines(gdf: GeoDataFrame, max_length: int, ignore_index=False) -> GeoDataFrame:
    """Cuts lines of a GeoDataFrame into pieces of a given length.

    Args:
        gdf: GeoDataFrame.
        max_length: The maximum length of the lines in the output GeoDataFrame.
        ignore_index: If True, the resulting axis will be labeled 0, 1, …, n - 1.
            Defaults to False.

    Returns:
        A GeoDataFrame with lines cut to the maximum distance.

    Note:
        This method is time consuming for large networks and low 'max_length'.

    Examples
    --------
    >>> from sgis import read_parquet_url, cut_lines
    >>> roads = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_oslo_2022.parquet")
    >>> roads.length.describe().round(1)
    count    93395.0
    mean        41.2
    std         78.5
    min          0.2
    25%         14.0
    50%         27.7
    75%         47.5
    max       5213.7
    dtype: float64
    >>> roads = cut_lines(roads, max_length=100)
    >>> roads.length.describe().round(1)
    count    126304.0
    mean         30.5
    std          30.1
    min           0.0
    25%           5.7
    50%          22.5
    75%          44.7
    max         100.0
    dtype: float64
    """
    gdf["geometry"] = force_2d(gdf.geometry)

    gdf = gdf.explode(ignore_index=ignore_index, index_parts=False)

    long_lines = gdf.loc[gdf.length > max_length]

    if not len(long_lines):
        return gdf

    for x in [10, 5, 1]:
        max_ = max(long_lines.length)
        while max_ > max_length * x + 1:
            max_ = max(long_lines.length)

            long_lines = cut_lines_once(long_lines, max_length)

            if max_ == max(long_lines.length):
                break

    long_lines = long_lines.explode(ignore_index=ignore_index, index_parts=False)

    short_lines = gdf.loc[gdf.length <= max_length]

    return pd.concat([short_lines, long_lines], ignore_index=ignore_index)


def cut_lines_once(
    lines: GeoDataFrame,
    distances: int | float | str | Series,
    ignore_index: bool = False,
) -> GeoDataFrame:
    """Cuts lines of a GeoDataFrame in two at the given distance or distances.

    Takes a GeoDataFrame of lines and cuts each line in two. If distances is a number,
    all lines will be cut at the same length.

    Args:
        gdf: GeoDataFrame.
        distances: The distance from the start of the lines to cut at. Either a number,
            the name of a column or array-like of same length as the line GeoDataFrame.
        ignore_index: If True, the resulting axis will be labeled 0, 1, …, n - 1.
            Defaults to False.

    Examples
    --------
    >>> from sgis import cut_lines_once, to_gdf
    >>> import pandas as pd
    >>> from shapely.geometry import LineString
    >>> gdf = to_gdf(LineString([(0, 0), (1, 1), (2, 2)]))
    >>> gdf = pd.concat([gdf, gdf, gdf])
    >>> gdf
                                                geometry
    0  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...
    0  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...
    0  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...

    Cut all lines at 1 unit from the start of the lines.

    >>> cut_lines_once(gdf, 1)
                                                geometry
    0      LINESTRING (0.00000 0.00000, 0.70711 0.70711)
    1  LINESTRING (0.70711 0.70711, 1.00000 1.00000, ...
    2      LINESTRING (0.00000 0.00000, 0.70711 0.70711)
    3  LINESTRING (0.70711 0.70711, 1.00000 1.00000, ...
    4      LINESTRING (0.00000 0.00000, 0.70711 0.70711)
    5  LINESTRING (0.70711 0.70711, 1.00000 1.00000, ...

    Cut distance as column.

    >>> gdf["dist"] = [1, 2, 3]
    >>> cut_lines_once(gdf, "dist")
                                                geometry  dist
    0      LINESTRING (0.00000 0.00000, 0.70711 0.70711)     1
    1  LINESTRING (0.70711 0.70711, 1.00000 1.00000, ...     1
    2  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...     2
    3      LINESTRING (1.41421 1.41421, 2.00000 2.00000)     2
    4  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...     3

    Cut distance as list (same result as above).

    >>> cut_lines_once(gdf, [1, 2, 3])
                                                geometry  dist
    0      LINESTRING (0.00000 0.00000, 0.70711 0.70711)     1
    1  LINESTRING (0.70711 0.70711, 1.00000 1.00000, ...     1
    2  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...     2
    3      LINESTRING (1.41421 1.41421, 2.00000 2.00000)     2
    4  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...     3
    """

    def _cut(line: LineString, distance: int | float) -> list[LineString]:
        """From the shapely docs"""
        if distance <= 0.0 or distance >= line.length:
            return line
        coords = list(line.coords)
        for i, p in enumerate(coords):
            prd = line.project(Point(p))
            if prd == distance:
                return [LineString(coords[: i + 1]), LineString(coords[i:])]
            if prd > distance:
                cp = line.interpolate(distance)
                return [
                    LineString(coords[:i] + [(cp.x, cp.y)]),
                    LineString([(cp.x, cp.y)] + coords[i:]),
                ]

    crs = lines.crs
    geom_col = lines._geometry_column_name

    lines = lines.copy()

    # cutting lines will give lists of linestrings in the geometry column. Ignoring
    # the warning it triggers
    warnings.filterwarnings(
        "ignore", message="Geometry column does not contain geometry."
    )

    if isinstance(distances, str):
        lines[geom_col] = np.vectorize(_cut)(lines[geom_col], lines[distances])
    else:
        lines[geom_col] = np.vectorize(_cut)(lines[geom_col], distances)

    # explode will give pandas df if not gdf is constructed
    lines = GeoDataFrame(
        lines.explode(ignore_index=ignore_index, index_parts=False),
        geometry=geom_col,
        crs=crs,
    )

    return lines


def make_node_ids(
    lines: GeoDataFrame,
    wkt: bool = True,
) -> tuple[GeoDataFrame, GeoDataFrame]:
    """Gives the lines unique node ids and returns lines (edges) and nodes.

    Takes the first and last point of each line and creates a GeoDataFrame of
    nodes (points) with a column 'node_id'. The node ids are then assigned to the
    input GeoDataFrame of lines as the columns 'source' and 'target'.

    Args:
        lines: GeoDataFrame with line geometries
        wkt: If True (default), the resulting nodes will include the column 'wkt',
            containing the well-known text representation of the geometry. If False, it
            will include the column 'coords', a tuple with x and y geometries.

    Returns:
        A tuple of two GeoDataFrames, one with the lines and one with the nodes.

    Note:
        The lines must be singlepart linestrings.
    """
    if wkt:
        lines = make_edge_wkt_cols(lines)
        geomcol1, geomcol2, geomcol_final = "source_wkt", "target_wkt", "wkt"
    else:
        lines = make_edge_coords_cols(lines)
        geomcol1, geomcol2, geomcol_final = "source_coords", "target_coords", "coords"

    # remove identical lines in opposite directions
    lines["meters_"] = lines.length.astype(str)

    sources = lines[[geomcol1, geomcol2, "meters_"]].rename(
        columns={geomcol1: geomcol_final, geomcol2: "temp"}
    )
    targets = lines[[geomcol1, geomcol2, "meters_"]].rename(
        columns={geomcol2: geomcol_final, geomcol1: "temp"}
    )

    nodes = (
        pd.concat([sources, targets], axis=0, ignore_index=True)
        .drop_duplicates([geomcol_final, "temp", "meters_"])
        .drop("meters_", axis=1)
    )

    lines = lines.drop("meters_", axis=1)

    nodes["n"] = nodes.assign(n=1).groupby(geomcol_final)["n"].transform("sum")

    nodes = nodes.drop_duplicates(subset=[geomcol_final]).reset_index(drop=True)

    nodes["node_id"] = nodes.index
    nodes["node_id"] = nodes["node_id"].astype(str)

    id_dict = {
        geom: node_id
        for geom, node_id in zip(nodes[geomcol_final], nodes["node_id"], strict=True)
    }
    lines["source"] = lines[geomcol1].map(id_dict)
    lines["target"] = lines[geomcol2].map(id_dict)

    n_dict = {geom: n for geom, n in zip(nodes[geomcol_final], nodes["n"], strict=True)}
    lines["n_source"] = lines[geomcol1].map(n_dict)
    lines["n_target"] = lines[geomcol2].map(n_dict)

    if wkt:
        nodes["geometry"] = gpd.GeoSeries.from_wkt(nodes[geomcol_final], crs=lines.crs)
    else:
        nodes["geometry"] = GeoSeries(
            [Point(geom) for geom in nodes[geomcol_final]], crs=lines.crs
        )
    nodes = gpd.GeoDataFrame(nodes, geometry="geometry", crs=lines.crs)
    nodes = nodes.reset_index(drop=True)

    lines = _push_geom_col(lines)

    return lines, nodes


def make_edge_coords_cols(lines: GeoDataFrame) -> GeoDataFrame:
    """Get the wkt of the first and last points of lines as columns.

    It takes a GeoDataFrame of LineStrings and returns a GeoDataFrame with two new
    columns, source_coords and target_coords, which are the x and y coordinates of the
    first and last points of the LineStrings in a tuple. The lines all have to be

    Args:
        lines (GeoDataFrame): the GeoDataFrame with the lines

    Returns:
        A GeoDataFrame with new columns 'source_coords' and 'target_coords'
    """
    try:
        lines, endpoints = _prepare_make_edge_cols_simple(lines)
    except ValueError:
        lines, endpoints = _prepare_make_edge_cols(lines)

    coords = [(geom.x, geom.y) for geom in endpoints.geometry]
    lines["source_coords"], lines["target_coords"] = (
        coords[0::2],
        coords[1::2],
    )

    return lines


def make_edge_wkt_cols(lines: GeoDataFrame) -> GeoDataFrame:
    """Get coordinate tuples of the first and last points of lines as columns.

    It takes a GeoDataFrame of LineStrings and returns a GeoDataFrame with two new
    columns, source_wkt and target_wkt, which are the WKT representations of the first
    and last points of the LineStrings

    Args:
        lines (GeoDataFrame): the GeoDataFrame with the lines

    Returns:
        A GeoDataFrame with new columns 'source_wkt' and 'target_wkt'
    """
    try:
        lines, endpoints = _prepare_make_edge_cols_simple(lines)
    except ValueError:
        lines, endpoints = _prepare_make_edge_cols(lines)

    wkt_geom = [
        f"POINT ({x} {y})" for x, y in zip(endpoints.x, endpoints.y, strict=True)
    ]
    lines["source_wkt"], lines["target_wkt"] = (
        wkt_geom[0::2],
        wkt_geom[1::2],
    )

    return lines


def close_network_holes(
    lines: GeoDataFrame,
    max_distance: int | float,
    max_angle: int,
    hole_col: str | None = "hole",
):
    """Fills network gaps with straigt lines.

    Fills holes in the network by connecting deadends with the nodes that are
    within the 'max_distance' distance.

    Args:
        lines: GeoDataFrame with lines.
        max_distance: The maximum distance for the holes to be filled.
        max_angle: Absolute number between 0 and 180 that represents the maximum
            difference in angle between the new line and the prior, i.e. the line
            at which the deadend terminates. A value of 0 means the new lines must
            have the exact same angle as the prior line, and 180 means the new
            lines can go in any direction.
        hole_col: If you want to keep track of which lines were added, you can add a
            column with a value of 1. Defaults to 'hole'

    Returns:
        The input GeoDataFrame with new lines added.

    Examples
    --------
    Read road data with small gaps.

    >>> import sgis as sg
    >>> roads = sg.read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_oslo_2022.parquet")

    Roads need to be singlepart linestrings for this to work.

    >>> from shapely import line_merge
    >>> roads.geometry = line_merge(roads.geometry)

    Fill gaps shorter than 1.1 meters.

    >>> filled = sg.close_network_holes(roads, max_distance=1.1, max_angle=180)
    >>> filled.hole.value_counts()
    Name: connected, dtype: int64
    0    93395
    1     7102
    Name: hole, dtype: int64

    Compare the number of isolated lines before and after.

    >>> roads = sg.get_largest_component(roads)
    >>> roads.connected.value_counts()
    1.0    85638
    0.0     7757
    Name: connected, dtype: int64

    >>> filled = sg.get_largest_component(filled)
    >>> filled.connected.value_counts()
    1.0    100315
    0.0       180
    Name: connected, dtype: int64

    Fill only gaps with an angle deviation between 0 and 30 compared to the prior line.

    >>> filled = sg.close_network_holes(roads, max_distance=1.1, max_angle=30)
    >>> filled.hole.value_counts()
    0    93395
    1     7092
    Name: hole, dtype: int64

    It's not always wise to fill gaps. In the case of this data, these small gaps are
    intentional. They are road blocks where most cars aren't allowed to pass. Fill the
    holes only if it makes the travel times/routes more realistic.
    """
    lines, nodes = make_node_ids(lines)

    new_lines = _find_holes_all_lines(
        lines,
        nodes,
        max_distance,
        max_angle=max_angle,
    )

    if not len(new_lines):
        lines[hole_col] = (
            0 if hole_col not in lines.columns else lines[hole_col].fillna(0)
        )
        return lines

    new_lines = make_edge_wkt_cols(new_lines)

    wkt_id_dict = {
        wkt: id for wkt, id in zip(nodes["wkt"], nodes["node_id"], strict=True)
    }
    new_lines["source"] = new_lines["source_wkt"].map(wkt_id_dict)
    new_lines["target"] = new_lines["target_wkt"].map(wkt_id_dict)

    if hole_col:
        new_lines[hole_col] = 1
        lines[hole_col] = (
            0 if hole_col not in lines.columns else lines[hole_col].fillna(0)
        )

    return pd.concat([lines, new_lines], ignore_index=True)


def close_network_holes_to_deadends(
    lines: GeoDataFrame,
    max_distance: int | float,
    hole_col: str | None = "hole",
):
    """Fills gaps between two deadends if the distance is less than 'max_distance'.

    Fills holes between deadends in the network with straight lines if the distance is
    less than 'max_distance'.

    Args:
        lines: GeoDataFrame with lines
        max_distance: The maximum distance between two nodes to be considered a hole.
        hole_col: If you want to keep track of which lines were added, you can add a
            column with a value of 1. Defaults to 'hole'

    Returns:
        The input GeoDataFrame with new lines added.

    Examples
    --------
    Read road data with small gaps.

    >>> import sgis as sg
    >>> roads = sg.read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_oslo_2022.parquet")

    Roads need to be singlepart linestrings for this to work.

    >>> from shapely import line_merge
    >>> roads.geometry = line_merge(roads.geometry)

    Check for number of isolated lines now.

    >>> roads = sg.get_largest_component(roads)
    >>> roads.connected.value_counts()
    1.0    85638
    0.0     7757
    Name: connected, dtype: int64

    Fill gaps shorter than 1.1 meters.

    >>> filled = sg.close_network_holes_to_deadends(roads, max_distance=1.1)
    >>> roads = sg.get_largest_component(roads)
    >>> roads.connected.value_counts()
    1.0    100315
    0.0       180
    Name: connected, dtype: int64

    It's not always wise to fill gaps. In the case of this data, these small gaps are
    intentional. They are road blocks where most cars aren't allowed to pass. Fill the
    holes only if it makes the travel times/routes more realistic.
    """
    lines, nodes = make_node_ids(lines)

    new_lines = _find_holes_deadends(nodes, max_distance)

    if not len(new_lines):
        lines[hole_col] = (
            0 if hole_col not in lines.columns else lines[hole_col].fillna(0)
        )
        return lines

    new_lines = make_edge_wkt_cols(new_lines)

    wkt_id_dict = {
        wkt: id for wkt, id in zip(nodes["wkt"], nodes["node_id"], strict=True)
    }
    new_lines["source"] = new_lines["source_wkt"].map(wkt_id_dict)
    new_lines["target"] = new_lines["target_wkt"].map(wkt_id_dict)

    if hole_col:
        new_lines[hole_col] = 1
        lines[hole_col] = (
            0 if hole_col not in lines.columns else lines[hole_col].fillna(0)
        )

    return pd.concat([lines, new_lines], ignore_index=True)


def _find_holes_all_lines(
    lines: GeoDataFrame,
    nodes: GeoDataFrame,
    max_distance: int,
    max_angle: int = 90,
):
    """Creates lines between deadends and closest node.

    Creates lines if distance is less than max_distance and angle less than max_angle.

    wkt: well-known text, e.g. "POINT (60 10)"
    """
    k = 50 if len(nodes) >= 50 else len(nodes)
    crs = nodes.crs

    # remove duplicates of lines going both directions
    lines["sorted"] = [
        "_".join(sorted([s, t]))
        for s, t in zip(lines["source"], lines["target"], strict=True)
    ]

    no_dups = lines.drop_duplicates("sorted")

    no_dups, nodes = make_node_ids(no_dups)

    # make point gdf for the deadends and the other endpoint of the deadend lines
    deadends_target = no_dups.loc[no_dups.n_target == 1].rename(
        columns={"target_wkt": "wkt", "source_wkt": "wkt_other_end"}
    )
    deadends_source = no_dups.loc[no_dups.n_source == 1].rename(
        columns={"source_wkt": "wkt", "target_wkt": "wkt_other_end"}
    )
    deadends = pd.concat([deadends_source, deadends_target], ignore_index=True)

    if len(deadends) <= 1:
        return []

    deadends_other_end = deadends.copy()
    deadends_other_end["geometry"] = gpd.GeoSeries.from_wkt(
        deadends_other_end["wkt_other_end"], crs=crs
    )

    deadends["geometry"] = gpd.GeoSeries.from_wkt(deadends["wkt"], crs=crs)

    deadends_array = coordinate_array(deadends)
    nodes_array = coordinate_array(nodes)

    all_dists, all_indices = k_nearest_neighbors(deadends_array, nodes_array, k=k)

    deadends_other_end_array = coordinate_array(deadends_other_end)

    def get_angle(array_a, array_b):
        dx = array_b[:, 0] - array_a[:, 0]
        dy = array_b[:, 1] - array_a[:, 1]

        angles_rad = np.arctan2(dx, dy)
        angles_degrees = np.degrees(angles_rad)
        return angles_degrees

    # now to find the lines that have the correct angle and distance
    # and endpoints of the new lines in lists, looping through the k neighbour points
    new_sources: list[str] = []
    new_targets: list[str] = []
    for i in np.arange(1, k):
        # to break out of the loop if no new_targets that meet the condition are found
        len_now = len(new_sources)

        # selecting the arrays for the current k neighbour
        indices = all_indices[:, i]
        dists = all_dists[:, i]

        these_nodes_array = coordinate_array(nodes.loc[indices])

        if np.all(deadends_other_end_array == these_nodes_array):
            continue

        angles_deadend_to_node = get_angle(deadends_array, these_nodes_array)

        angles_deadend_to_deadend_other_end = get_angle(
            deadends_other_end_array, deadends_array
        )

        angles_difference = np.abs(
            np.abs(angles_deadend_to_deadend_other_end) - np.abs(angles_deadend_to_node)
        )

        angles_difference[
            np.all(deadends_other_end_array == these_nodes_array, axis=1)
        ] = np.nan

        condition = (dists <= max_distance) & (angles_difference <= max_angle)

        from_wkt = deadends.loc[condition, "wkt"]
        to_idx = indices[condition]
        to_wkt = nodes.loc[to_idx, "wkt"]

        # now add the wkts to the lists of new sources and targets. If the source
        # is already added, the new wks will not be added again
        new_targets = new_targets + [
            t for f, t in zip(from_wkt, to_wkt, strict=True) if f not in new_sources
        ]
        new_sources = new_sources + [
            f for f, _ in zip(from_wkt, to_wkt, strict=True) if f not in new_sources
        ]

        # break out of the loop when no new new_targets meet the condition
        if len_now == len(new_sources):
            break

    # make GeoDataFrame with straight lines
    new_sources = gpd.GeoSeries.from_wkt(new_sources, crs=crs)
    new_targets = gpd.GeoSeries.from_wkt(new_targets, crs=crs)
    new_lines = shortest_line(new_sources, new_targets)
    new_lines = gpd.GeoDataFrame({"geometry": new_lines}, geometry="geometry", crs=crs)

    if not len(new_lines):
        return new_lines

    new_lines = make_edge_wkt_cols(new_lines)

    return new_lines


def _find_holes_deadends(nodes: GeoDataFrame, max_distance: float | int):
    """Creates lines between two deadends if between max_distance and min_dist.

    It takes a GeoDataFrame of nodes, chooses the deadends, and creates a straight line
    between the closest deadends if the distance is no greater than 'max_distance'.

    Args:
        nodes: the nodes of the network
        max_distance: The maximum distance between two nodes to be connected.

    Returns:
        A GeoDataFrame with the new lines.
    """
    crs = nodes.crs

    # deadends are nodes that appear only once
    deadends = nodes[nodes["n"] == 1]

    # have to reset index to be able to integrate with numpy/scikit-learn
    deadends = deadends.reset_index(drop=True)

    if len(deadends) <= 1:
        return []

    deadends_array = coordinate_array(deadends)

    dists, indices = k_nearest_neighbors(deadends_array, deadends_array, k=2)

    # choose the second column (the closest neighbour)
    dists = dists[:, 1]
    indices = indices[:, 1]

    # get the geometry of the distances no greater than max_distance
    # the from geometries can be taken directly from the deadends index,
    # since 'dists' has the same index. 'to_geom' must be selected through the index
    # of the neighbours ('indices')
    condition = dists < max_distance
    from_geom = deadends.loc[condition, "geometry"].reset_index(drop=True)
    to_idx = indices[condition]
    to_geom = deadends.loc[to_idx, "geometry"].reset_index(drop=True)

    # GeoDataFrame with straight lines
    new_lines = shortest_line(from_geom, to_geom)
    new_lines = gpd.GeoDataFrame({"geometry": new_lines}, geometry="geometry", crs=crs)

    if not len(new_lines):
        return new_lines

    return new_lines


def _prepare_make_edge_cols(
    lines: GeoDataFrame,
) -> tuple[GeoDataFrame, GeoDataFrame]:
    lines = lines.loc[lines.geom_type != "LinearRing"]

    if not all(lines.geom_type == "LineString"):
        multilinestring_error_message = (
            "MultiLineStrings have more than two endpoints. "
            "Try shapely.line_merge and/or explode() to get LineStrings. "
            "Or use the Network class methods, where the lines are prepared correctly."
        )
        if any(lines.geom_type == "MultiLinestring"):
            raise ValueError(multilinestring_error_message)
        else:
            raise ValueError(
                "You have mixed geometries. Only lines are accepted. "
                "Try using: to_single_geom_type(gdf, 'lines')."
            )

    geom_col = lines._geometry_column_name

    # some LinearRings are coded as LineStrings and need to be removed manually
    boundary = lines[geom_col].boundary
    circles = boundary.loc[boundary.is_empty]
    lines = lines[~lines.index.isin(circles.index)]

    endpoints = lines[geom_col].boundary.explode(ignore_index=True)

    if len(endpoints) / len(lines) != 2:
        raise ValueError(
            "The lines should have only two endpoints each. "
            "Try splitting multilinestrings with explode."
        )

    return lines, endpoints


def _prepare_make_edge_cols_simple(
    lines: GeoDataFrame,
) -> tuple[GeoDataFrame, GeoDataFrame]:
    """Faster version of _prepare_make_edge_cols."""

    endpoints = lines[lines._geometry_column_name].boundary.explode(ignore_index=True)

    if len(endpoints) / len(lines) != 2:
        raise ValueError(
            "The lines should have only two endpoints each. "
            "Try splitting multilinestrings with explode."
        )

    return lines, endpoints


def _roundabouts_to_intersections(roads, query="ROADTYPE=='Rundkjøring'"):
    from shapely.geometry import LineString
    from shapely.ops import nearest_points

    from .buffer_dissolve_explode import buffdissexp

    roundabouts = roads.query(query)
    not_roundabouts = roads.loc[~roads.index.isin(roundabouts.index)]

    roundabouts = buffdissexp(roundabouts[["geometry"]], 1)
    roundabouts["roundidx"] = roundabouts.index

    border_to = roundabouts.overlay(not_roundabouts, how="intersection")

    # for hver rundkjøring: lag linjer mellom rundkjøringens mitdpunkt og hver linje
    # som grenser til rundkjøringen
    as_intersections = []
    for idx in roundabouts.roundidx:
        this = roundabouts.loc[roundabouts.roundidx == idx]
        border_to_this = border_to.loc[border_to.roundidx == idx].drop(
            "roundidx", axis=1
        )

        midpoint = this.unary_union.centroid

        # straight lines to the midpoint
        for i, line in enumerate(border_to_this.geometry):
            closest_point = nearest_points(midpoint, line)[1]
            border_to_this.geometry.iloc[i] = LineString([closest_point, midpoint])

        as_intersections.append(border_to_this)

    as_intersections = GeoDataFrame(
        pd.concat(as_intersections, ignore_index=True), crs=roads.crs
    )
    out = GeoDataFrame(
        pd.concat([not_roundabouts, as_intersections], ignore_index=True), crs=roads.crs
    )

    return out



"""Interactive map of one or more GeoDataFrames with layers that can be toggles on/off.

This module holds the Explore class, which is the basis for the explore, samplemap and
clipmap functions from the 'maps' module.
"""
import warnings
from statistics import mean

import branca as bc
import folium
import matplotlib
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from shapely.geometry import LineString

from ..geopandas_tools.general import clean_geoms, random_points_in_polygons
from ..geopandas_tools.geometry_types import get_geom_type
from .map import Map


# the geopandas._explore raises a deprication warning. Ignoring for now.
warnings.filterwarnings(
    action="ignore", category=matplotlib.MatplotlibDeprecationWarning
)
pd.options.mode.chained_assignment = None


# gray for NaNs
NAN_COLOR = "#969696"


# cols to not show when hovering over geometries (tooltip)
COLS_TO_DROP = ["color", "col_as_int", "geometry"]


# from geopandas
_MAP_KWARGS = [
    "location",
    "prefer_canvas",
    "no_touch",
    "disable_3d",
    "png_enabled",
    "zoom_control",
    "crs",
    "zoom_start",
    "left",
    "top",
    "position",
    "min_zoom",
    "max_zoom",
    "min_lat",
    "max_lat",
    "min_lon",
    "max_lon",
    "max_bounds",
]


class Explore(Map):
    def __init__(
        self,
        *gdfs,
        column: str | None = None,
        popup: bool = True,
        max_zoom: int = 30,
        show_in_browser: bool = False,
        **kwargs,
    ):
        super().__init__(*gdfs, column=column, **kwargs)
        self.popup = popup
        self.max_zoom = max_zoom
        self.show_in_browser = show_in_browser

        self._to_single_geom_type()

        if self._is_categorical:
            if len(self.gdfs) == 1:
                self._split_categories()
        else:
            if not self._cmap:
                self._cmap = "viridis"
            self.cmap_start = kwargs.pop("cmap_start", 0)
            self.cmap_stop = kwargs.pop("cmap_stop", 256)

    def explore(self, column: str | None = None, **kwargs) -> None:
        if column:
            self._column = column
            self._update_column()
            kwargs.pop("column", None)
        self.to_show = self._gdfs
        self._explore(**kwargs)

    def samplemap(
        self,
        size: int = 1000,
        column: str | None = None,
        sample_from_first: bool = True,
        **kwargs,
    ) -> None:
        if column:
            self._column = column
            self._update_column()
            kwargs.pop("column", None)

        self.previous_sample_count = 0

        if sample_from_first:
            sample = self._gdfs[0].sample(1)
        else:
            sample = self._gdf.sample(1)

        # convert lines to polygons
        if get_geom_type(sample) == "line":
            sample["geometry"] = sample.buffer(1)

        if get_geom_type(sample) == "polygon":
            random_point = random_points_in_polygons(sample, 1)

        # if point or mixed geometries
        else:
            random_point = sample.centroid

        gdfs: tuple[GeoDataFrame] = ()
        for gdf in self._gdfs:
            gdf = gdf.clip(random_point.buffer(size))
            gdfs = gdfs + (gdf,)
        self._gdfs = gdfs
        self._gdf = pd.concat(gdfs, ignore_index=True)
        self._explore(**kwargs)

    def clipmap(
        self,
        mask,
        column: str | None = None,
        **kwargs,
    ) -> None:
        if column:
            self._column = column
            self._update_column()
            kwargs.pop("column", None)

        gdfs: tuple[GeoDataFrame] = ()
        for gdf in self._gdfs:
            gdf = gdf.clip(mask)
            collections = gdf.loc[gdf.geom_type == "GeometryCollection"]
            if len(collections):
                collections = collections.explode(index_parts=False)
                gdf = pd.concat([gdf, collections], ignore_index=False)
            gdfs = gdfs + (gdf,)
        self._gdfs = gdfs
        self._gdf = pd.concat(gdfs, ignore_index=True)
        self._explore(**kwargs)

    def _explore(self, **kwargs):
        self.kwargs = self.kwargs | kwargs

        if self._is_categorical:
            self._create_categorical_map()
        else:
            self._create_continous_map()

        if self.show_in_browser:
            self.map.show_in_browser()
        else:
            display(self.map)

    def _split_categories(self):
        new_gdfs, new_labels = [], []
        for cat in self._unique_values:
            gdf = self.gdf.loc[self.gdf[self.column] == cat]
            new_gdfs.append(gdf)
            new_labels.append(cat)
        self._gdfs = new_gdfs
        self._gdf = pd.concat(new_gdfs, ignore_index=True)
        self.labels = new_labels

    def _to_single_geom_type(self):
        """Buffer gdf if mixed geometry types. Expode to singlepart.

        Because Folium does not handle mixed geometries well.
        """

        new_gdfs = []
        for gdf in self._gdfs:
            if get_geom_type(gdf) == "mixed":
                gdf[gdf._geometry_column_name] = gdf.buffer(0.1)
            gdf = gdf.explode(index_parts=False)
            new_gdfs.append(gdf)
        self._gdfs = new_gdfs
        self._gdf = pd.concat(new_gdfs, ignore_index=True)
        self._nan_idx = self._gdf[self._column].isna()

    def _update_column(self):
        self._is_categorical = self._check_if_categorical()
        self._fill_missings()
        self._gdf = pd.concat(self._gdfs, ignore_index=True)

    def _create_categorical_map(self):
        self._get_categorical_colors()

        self.map = self._explore_return(
            self._gdf,
            return_="empty_map",
            max_zoom=self.max_zoom,
            popup=self.popup,
            **self.kwargs,
        )

        for gdf, label in zip(self._gdfs, self.labels, strict=True):
            if not len(gdf):
                continue
            f = folium.FeatureGroup(name=label)

            gjs = self._explore_return(
                gdf,
                color=gdf["color"],
                return_="geojson",
                tooltip=self._tooltip_cols(gdf),
                **{
                    key: value
                    for key, value in self.kwargs.items()
                    if key not in ["title"]
                },
            )
            f.add_child(gjs)
            self.map.add_child(f)
        _categorical_legend(
            self.map,
            self._column,
            self._categories_colors_dict.keys(),
            self._categories_colors_dict.values(),
        )
        folium.TileLayer("stamentoner").add_to(self.map)
        folium.TileLayer("cartodbdark_matter").add_to(self.map)
        self.map.add_child(folium.LayerControl())

    def _create_continous_map(self):
        self._prepare_continous_map()
        if self.scheme:
            classified = self._classify_from_bins(self._gdf, bins=self.bins)
            classified_sequential = self._push_classification(classified)
            n_colors = len(np.unique(classified_sequential)) - any(self._nan_idx)
            unique_colors = self._get_continous_colors(n=n_colors)

        self.map = self._explore_return(
            self._gdf,
            return_="empty_map",
            max_zoom=self.max_zoom,
            popup=self.popup,
            **self.kwargs,
        )

        colorbar = bc.colormap.StepColormap(
            unique_colors,
            vmin=self._gdf[self._column].min(),
            vmax=self._gdf[self._column].max(),
            caption=self._column,
            index=self.bins,
            #  **colormap_kwds,
        )

        for gdf, label in zip(self._gdfs, self.labels, strict=True):
            if not len(gdf):
                continue
            f = folium.FeatureGroup(name=label)

            classified = self._classify_from_bins(gdf, bins=self.bins)
            colorarray = unique_colors[classified]

            gjs = self._explore_return(
                gdf,
                tooltip=self._tooltip_cols(gdf),
                color=colorarray,
                return_="geojson",
                **{key: value for key, value in self.kwargs.items() if key != "title"},
            )
            f.add_child(gjs)
            self.map.add_child(f)

        self.map.add_child(colorbar)
        folium.TileLayer("stamentoner").add_to(self.map)
        folium.TileLayer("cartodbdark_matter").add_to(self.map)
        self.map.add_child(folium.LayerControl())

    def _tooltip_cols(self, gdf: GeoDataFrame) -> list:
        if "tooltip" in self.kwargs:
            tooltip = self.kwargs.pop("tooltip")
            return tooltip
        return [col for col in gdf.columns if col not in COLS_TO_DROP]

    def _explore_return(
        self,
        df,
        return_: str,
        column=None,
        color=None,
        attr=None,
        tiles="OpenStreetMap",
        tooltip=True,
        popup=False,
        highlight=True,
        width="100%",
        height="100%",
        control_scale=True,
        marker_type=None,
        marker_kwds={},
        style_kwds={},
        highlight_kwds={},
        tooltip_kwds={},
        popup_kwds={},
        map_kwds={},
        **kwargs,
    ):
        """Contains the nessecary parts of the geopandas _explore function.

        Also has a return_ parameter that controls what is returned. This should be
        replaced by separate functions, and irrelevant parameters should be removed.
        """
        # xyservices is an optional dependency
        try:
            import xyzservices

            has_xyzservices = True
        except ImportError:
            has_xyzservices = False

        gdf = df.copy()

        # convert LinearRing to LineString
        rings_mask = df.geom_type == "LinearRing"
        if rings_mask.any():
            gdf.geometry[rings_mask] = gdf.geometry[rings_mask].apply(
                lambda g: LineString(g)
            )

        if gdf.crs is None:
            kwargs["crs"] = "Simple"
            tiles = None
        elif not gdf.crs.equals(4326):
            gdf = gdf.to_crs(4326)

        # create folium.Map object
        # Get bounds to specify location and map extent
        bounds = gdf.total_bounds
        location = kwargs.pop("location", None)
        if location is None:
            x = mean([bounds[0], bounds[2]])
            y = mean([bounds[1], bounds[3]])
            location = (y, x)
            if "zoom_start" in kwargs.keys():
                fit = False
            else:
                fit = True
        else:
            fit = False

        # get a subset of kwargs to be passed to folium.Map
        for i in _MAP_KWARGS:
            if i in map_kwds:
                raise ValueError(
                    f"'{i}' cannot be specified in 'map_kwds'. "
                    f"Use the '{i}={map_kwds[i]}' argument instead."
                )
        map_kwds = {
            **map_kwds,
            **{i: kwargs[i] for i in kwargs.keys() if i in _MAP_KWARGS},
        }

        if has_xyzservices:
            # match provider name string to xyzservices.TileProvider
            if isinstance(tiles, str):
                try:
                    tiles = xyzservices.providers.query_name(tiles)
                except ValueError:
                    pass

            if isinstance(tiles, xyzservices.TileProvider):
                attr = attr if attr else tiles.html_attribution
                map_kwds["min_zoom"] = tiles.get("min_zoom", 0)
                map_kwds["max_zoom"] = tiles.get("max_zoom", 30)
                tiles = tiles.build_url(scale_factor="{r}")

            m = folium.Map(
                location=location,
                control_scale=control_scale,
                tiles=tiles,
                attr=attr,
                width=width,
                height=height,
                **map_kwds,
            )

            # fit bounds to get a proper zoom level
            if fit:
                m.fit_bounds([[bounds[1], bounds[0]], [bounds[3], bounds[2]]])

        for map_kwd in _MAP_KWARGS:
            kwargs.pop(map_kwd, None)

        # set default style
        if "fillOpacity" not in style_kwds:
            style_kwds["fillOpacity"] = 0.5
        if "weight" not in style_kwds:
            style_kwds["weight"] = 2
        if "style_function" in style_kwds:
            style_kwds_function = style_kwds["style_function"]
            if not callable(style_kwds_function):
                raise ValueError("'style_function' has to be a callable")
            style_kwds.pop("style_function")
        else:

            def _no_style(x):
                return {}

            style_kwds_function = _no_style

        gdf["__folium_color"] = color

        stroke_color = style_kwds.pop("color", None)
        if not stroke_color:

            def _style_column(x):
                base_style = {
                    "fillColor": x["properties"]["__folium_color"],
                    "color": x["properties"]["__folium_color"],
                    **style_kwds,
                }
                return {
                    **base_style,
                    **style_kwds_function(x),
                }

            style_function = _style_column

        if highlight:
            if "fillOpacity" not in highlight_kwds:
                highlight_kwds["fillOpacity"] = 0.75

            def _style_highlight(x):
                return {**highlight_kwds}

            highlight_function = _style_highlight
        else:
            highlight_function = None

        # define default for points
        if marker_type is None:
            marker_type = "circle_marker"

        marker = marker_type
        if isinstance(marker_type, str):
            if marker_type == "marker":
                marker = folium.Marker(**marker_kwds)
            elif marker_type == "circle":
                marker = folium.Circle(**marker_kwds)
            elif marker_type == "circle_marker":
                marker_kwds["radius"] = marker_kwds.get("radius", 2)
                marker_kwds["fill"] = marker_kwds.get("fill", True)
                marker = folium.CircleMarker(**marker_kwds)
            else:
                raise ValueError(
                    "Only 'marker', 'circle', and 'circle_marker' are "
                    "supported as marker values"
                )

        gdf = clean_geoms(gdf)

        # prepare tooltip and popup
        if isinstance(gdf, GeoDataFrame):
            # add named index to the tooltip
            if gdf.index.name is not None:
                gdf = gdf.reset_index()
            # specify fields to show in the tooltip
            tooltip = _tooltip_popup("tooltip", tooltip, gdf, **tooltip_kwds)
            popup = _tooltip_popup("popup", popup, gdf, **popup_kwds)
        else:
            tooltip = None
            popup = None

        if "geojson" in return_:
            # add dataframe to map
            gjs = folium.GeoJson(
                gdf.__geo_interface__,
                tooltip=tooltip,
                popup=popup,
                marker=marker,
                style_function=style_function,
                highlight_function=highlight_function,
                **kwargs,
            )
            return gjs

        return m


def _tooltip_popup(type, fields, gdf, **kwds):
    """get tooltip or popup"""
    import folium

    # specify fields to show in the tooltip
    if fields is False or fields is None or fields == 0:
        return None
    else:
        if fields is True:
            fields = gdf.columns.drop(gdf.geometry.name).to_list()
        elif isinstance(fields, int):
            fields = gdf.columns.drop(gdf.geometry.name).to_list()[:fields]
        elif isinstance(fields, str):
            fields = [fields]

    for field in ["__plottable_column", "__folium_color"]:
        if field in fields:
            fields.remove(field)

    # Cast fields to str
    fields = list(map(str, fields))
    if type == "tooltip":
        return folium.GeoJsonTooltip(fields, **kwds)
    elif type == "popup":
        return folium.GeoJsonPopup(fields, **kwds)


def _categorical_legend(m, title, categories, colors):
    """
    Add categorical legend to a map

    The implementation is using the code originally written by Michel Metran
    (@michelmetran) and released on GitHub
    (https://github.com/michelmetran/package_folium) under MIT license.

    Copyright (c) 2020 Michel Metran

    Parameters
    ----------
    m : folium.Map
        Existing map instance on which to draw the plot
    title : str
        title of the legend (e.g. column name)
    categories : list-like
        list of categories
    colors : list-like
        list of colors (in the same order as categories)
    """

    # Header to Add
    head = """
    {% macro header(this, kwargs) %}
    <script src="https://code.jquery.com/ui/1.12.1/jquery-ui.js"></script>
    <script>$( function() {
        $( ".maplegend" ).draggable({
            start: function (event, ui) {
                $(this).css({
                    right: "auto",
                    top: "auto",
                    bottom: "auto"
                });
            }
        });
    });
    </script>
    <style type='text/css'>
      .maplegend {
        position: absolute;
        z-index:9999;
        background-color: rgba(255, 255, 255, .8);
        border-radius: 5px;
        box-shadow: 0 0 15px rgba(0,0,0,0.2);
        padding: 10px;
        font: 12px/14px Arial, Helvetica, sans-serif;
        right: 10px;
        bottom: 20px;
      }
      .maplegend .legend-title {
        text-align: left;
        margin-bottom: 5px;
        font-weight: bold;
        }
      .maplegend .legend-scale ul {
        margin: 0;
        margin-bottom: 0px;
        padding: 0;
        float: left;
        list-style: none;
        }
      .maplegend .legend-scale ul li {
        list-style: none;
        margin-left: 0;
        line-height: 16px;
        margin-bottom: 2px;
        }
      .maplegend ul.legend-labels li span {
        display: block;
        float: left;
        height: 14px;
        width: 14px;
        margin-right: 5px;
        margin-left: 0;
        border: 0px solid #ccc;
        }
      .maplegend .legend-source {
        color: #777;
        clear: both;
        }
      .maplegend a {
        color: #777;
        }
    </style>
    {% endmacro %}
    """
    import branca as bc

    # Add CSS (on Header)
    macro = bc.element.MacroElement()
    macro._template = bc.element.Template(head)
    m.get_root().add_child(macro)

    body = f"""
    <div id='maplegend {title}' class='maplegend'>
        <div class='legend-title'>{title}</div>
        <div class='legend-scale'>
            <ul class='legend-labels'>"""

    # Loop Categories
    for label, color in zip(categories, colors, strict=True):
        body += f"""
                <li><span style='background:{color}'></span>{label}</li>"""

    body += """
            </ul>
        </div>
    </div>
    """

    # Add Body
    body = bc.element.Element(body, "legend")
    m.get_root().html.add_child(body)



"""Network class with methods for manipulating line geometries.

The module includes functions for cutting and splitting lines, filling holes in the
network, finding and removing isolated network islands and creating unique node ids.
"""

import warnings
from copy import copy, deepcopy

import numpy as np
from geopandas import GeoDataFrame
from pandas import DataFrame
from shapely import line_merge

from ..exceptions import ZeroLinesError
from ..geopandas_tools.general import clean_geoms
from ..geopandas_tools.geometry_types import to_single_geom_type
from ..geopandas_tools.line_operations import (
    close_network_holes,
    close_network_holes_to_deadends,
    cut_lines,
    get_component_size,
    get_largest_component,
    make_node_ids,
)


class Network:
    """Prepares a GeoDataFrame of lines for network analysis.

    The class can be used as the 'network' parameter in the NetworkAnalysis class for
    undirected network analysis. For directed network analysis, use the DirectedNetwork
    class.

    The Network class also contains methods for optimizing the network. The
    remove_isolated method will remove isolated network islands. This often
    dramatically increases the success rate of the network analyses. The network
    islands can also be inspected with the get_largest_component or
    get_component_size methods.

    If the network contains holes/gaps that are not supposed to be there, the
    close_network_holes method can be used. This will fill any network holes
    shorter than a given distance. Often, this should be run before removing the
    isolated network components.

    Long lines can be cut into equal length pieces with the cut_lines method.
    This is mainly relevant for service_area analysis, since shorter lines
    will give more accurate results here.

    All methods can be chained together like in pandas. However, all methods also
    overwrite the prevous object to save memory. Take a copy with the 'copy' or
    'deepcopy' methods before using another method if you want to save the previous
    instance.

    The 'source' and 'target' ids are also stored as points in the 'nodes' attribute,
    which is always kept up to date with the lines and the actual geometries. The ids
    therefore changes whenever the lines change, so they cannot be used as fixed
    identifiers.

    Args:
        gdf: a GeoDataFrame of line geometries.
        merge_lines: if True (default), multilinestrings within the same row will
            be merged if they overlap. if False, multilines will be split into
            separate rows of singlepart lines.
        allow_degree_units: If False (default), it will raise an exception if
            the coordinate reference system of 'gdf' is in degree units, i.e.
            unprojected (4326). If set to True, all crs are allowed, but it might
            raise exceptions and give erronous results.

    Attributes:
        gdf: the GeoDataFrame of lines with source and target ids.
        nodes: GeoDataFrame of points with unique ids. The ids are dynamic, and changes
            every time the lines change. The nodes cannot be changed directly.

    See also
    --------
    DirectedNetwork: subclass of Network for directed network analysis.

    Examples
    --------
    Read example data.

    >>> from sgis import read_parquet_url
    >>> roads = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_oslo_2022.parquet")
    >>> nw = Network(roads)
    >>> nw
    Network(3851 km, undirected)

    Check for isolated network islands.

    >>> nw = nw.get_largest_component()
    >>> nw.gdf.connected.value_counts()
    1.0    85638
    0.0     7757
    Name: connected, dtype: int64

    Remove the network islands.

    >>> nw = nw.remove_isolated()
    >>> nw.gdf.connected.value_counts()
    1.0    85638
    Name: connected, dtype: int64

    Filling small gaps/holes in the network.

    >>> len(nw.gdf)
    85638
    >>> nw = nw.close_network_holes(max_distance=1.5, max_angle=90, fillna=0)
    >>> len(nw.gdf)
    86929

    Cutting long lines into pieces. This is only relevant for service area analysis and
    similar analyses.

    >>> nw.gdf.length.max()
    5213.749177803526
    >>> nw = nw.cut_lines(100)
    >>> nw.gdf.length.max()
    100.00000000046512
    """

    def __init__(
        self,
        gdf: GeoDataFrame,
        *,
        allow_degree_units: bool = False,
    ):
        """The lines are fixed, welded together rowwise and exploded. Creates node-ids.

        Raises:
            TypeError: If 'gdf' is not of type GeoDataFrame.
            ZeroLinesError: If 'gdf' has zero rows.
            ValueError: If the coordinate reference system is in degree units and
                allow_degree_units is False.
        """
        if not isinstance(gdf, GeoDataFrame):
            raise TypeError(f"'lines' should be GeoDataFrame, got {type(gdf)}")

        if not len(gdf):
            raise ZeroLinesError

        if not allow_degree_units and gdf.crs.axis_info[0].unit_name == "degree":
            raise ValueError(
                "The crs cannot have degrees as unit. Change to a projected crs with "
                "e.g. 'metre' as unit. If you really want to use an unprojected crs, "
                "set 'allow_degree_units' to True."
            )

        self.gdf = self._prepare_network(gdf)

        self._make_node_ids()

        # attributes for the log
        if "connected" in self.gdf.columns:
            if all(self.gdf["connected"] == 1):
                self._isolated_removed = True
        else:
            self._isolated_removed = False

        self._percent_bidirectional = self._check_percent_bidirectional()

        # for the base Network class, the graph will be undirected in network analysis
        self._as_directed = False

    def remove_isolated(self):
        """Removes lines not connected to the largest network component.

        It creates a graph and finds the edges that are part of the largest
        connected component of the network. Then removes all lines not part of
        the largest component. Updates node ids if neccessary.

        Returns:
            Self

        See also:
            get_largest_component: to find the isolated lines without removing them
            get_component_size: to get the exact number of lines in the component

        Examples
        --------
        Compare the number of rows.

        >>> from sgis import Network, read_parquet_url
        >>> roads = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_oslo_2022.parquet")
        >>> nw = Network(roads)
        >>> len(nw.gdf)
        93395
        >>> nw = nw.remove_isolated()
        >>> len(nw.gdf)
        85638

        Removing isolated can have a large inpact on the number of missing values in
        network analyses.

        >>> nwa = NetworkAnalysis(
        ...     network=Network(roads),
        ...     rules=NetworkAnalysisRules(weight="meters")
        ... )
        >>> od = nwa.od_cost_matrix(points, points)
        >>> nwa.network = nwa.network.remove_isolated()
        >>> od = nwa.od_cost_matrix(points, points)
        >>> nwa.log[['isolated_removed', 'percent_missing']]
        isolated_removed  percent_missing
        0             False          23.2466
        1              True           0.3994
        """
        if not self._nodes_are_up_to_date():
            self._make_node_ids()

        self.gdf = get_largest_component(self.gdf)

        self.gdf = self.gdf.loc[self.gdf.connected == 1]

        self._isolated_removed = True

        return self

    def get_largest_component(self):
        """Create column 'connected' in the network.gdf, where '1' means connected.

        It takes the lines of the network, creates a graph, finds the largest component,
        and maps this as the value '1' in the column 'connected' and gives the rest the
        value 0.

        Returns:
            self

        See also:
            remove_isolated: to remove the isolated lines from the network
            get_component_size: to get the exact number of lines in the component

        Examples
        --------
        >>> nw = Network(roads)
        >>> nw = nw.get_largest_component()
        >>> nw.gdf.connected.value_counts()
        1.0    85638
        0.0     7757
        Name: connected, dtype: int64
        """
        self._update_nodes_if()
        if "connected" in self.gdf.columns:
            warnings.warn(
                "There is already a column 'connected' in the network. Run "
                ".remove_isolated() if you want to remove the isolated networks."
                "And write nw.gdf.connected... if you want to access the column",
                stacklevel=2,
            )
        self.gdf = get_largest_component(self.gdf)

        return self

    def get_component_size(self):
        """Finds the size of each component in the network.

        Creates the column "component_size" in the 'gdf' of the network, which
        indicates the number of lines in the network component the line is a part of.

        Returns:
            self

        See also:
            get_largest_component: to find the isolated lines without removing them
            get_component_size: to get the exact number of lines in the component

        Examples
        --------
        >>> nw = Network(roads)
        >>> nw = nw.get_component_size()
        >>> nw.gdf.component_size.value_counts().head()
        79180    85638
        2         1601
        4          688
        6          406
        3          346
        Name: component_size, dtype: int64
        """
        self._update_nodes_if()
        self.gdf = get_component_size(self.gdf)
        return self

    def close_network_holes(
        self,
        max_distance: int | float,
        max_angle: int,
        fillna: float | int | None,
        hole_col: str = "hole",
    ):
        """Fills network gaps with straigt lines.

        Fills holes in the network by connecting deadends with the the closest node
        that is within the 'max_distance' distance and have an angle less 'max_angle'.

        Args:
            max_distance: The maximum distance for the holes to be filled.
            max_angle: Absolute number between 0 and 180 that represents the maximum
                difference in angle between the new line and the prior, i.e. the line
                at which the deadend ends. A value of 0 means the new lines must
                have the exact same angle as the prior line, and 180 means the new
                lines can go in any direction.
            fillna: Numeric value to assign to all NaN values of the holes. This is
                chiefly the 'weight' column that is to be used in network analysis.
            hole_col: Holes will get the value 1 in a column named 'hole' by default,
                or what is specified as hole_col. If set to None or False, no column
                will be added.

        Returns:
            The input GeoDataFrame with new lines added.

        Note:
            The 'fillna' parameter has no default value, but can be set to None if NaN
            values should be preserved. This will, however, mean that the filled holes
            will be opened again in network analysis.

        Examples
        --------
        Check for number of isolated lines before filling gaps.

        >>> nw = Network(roads)
        >>> nw = nw.get_largest_component()
        >>> nw.gdf.connected.value_counts()
        1.0    85638
        0.0     7757
        Name: connected, dtype: int64

        Fill gaps shorter than 1.1 meters and an angle deviation of no more than
        30 degrees.

        >>> nw = nw.close_network_holes(max_distance=1.1, max_angle=30, fillna=0)
        >>> nw = nw.get_largest_component()
        >>> nw.gdf.connected.value_counts()
        1.0    100315
        0.0       191
        Name: connected, dtype: int64

        Fill gaps with all angles allowed.

        >>> nw = nw.close_network_holes(max_distance=1.1, max_angle=180, fillna=0)
        >>> nw = nw.get_largest_component()
        >>> nw.gdf.connected.value_counts()
        1.0    100315
        0.0       180
        Name: connected, dtype: int64

        It's not always wise to fill gaps. In the case of this data, these small gaps
        are intentional. They are road blocks where most cars aren't allowed to pass.
        Fill the holes only if it makes the travel times/routes more realistic.
        """
        self.gdf = close_network_holes(
            self.gdf,
            max_distance,
            hole_col=hole_col,
            max_angle=max_angle,
        )

        if fillna is not None:
            self.gdf.loc[self.gdf[hole_col] == 1] = self.gdf.loc[
                self.gdf[hole_col] == 1
            ].fillna(fillna)

        self._hole_col = hole_col

        return self

    def close_network_holes_to_deadends(
        self,
        max_distance: int | float,
        fillna: int | float | None,
        hole_col: str = "hole",
    ):
        """Fills holes between two deadends if the distance is less than the max_distance.

        It fills gaps in the network by finding the nearest neighbors of each node,
        then creating straigt lines between the nodes that are within the 'max_distance'
        of each other.

        Args:
            max_distance: The maximum distance between two nodes to be considered a hole.
            fillna: The value to give the closed holes in all columns with NaN. This
                has to be set in order for the closed holes to be included in network
                analyses. If set to 0, the lines will get a weight of 0.
            hole_col: Holes will get the value 1 in a column named 'hole' by default,
                or what is specified as hole_col. If set to None or False, no column
                will be added.

        Returns:
            The input GeoDataFrame with new lines added.

        Note:
            The 'fillna' parameter has no default value, but can be set to None if NaN
            values should be preserved. This will, however, mean that the filled holes
            will be opened again in network analysis.

        Examples
        --------
        Check for number of isolated lines before filling gaps.

        >>> nw = Network(roads)
        >>> nw = nw.get_largest_component()
        >>> nw.gdf.connected.value_counts()
        1.0    85638
        0.0     7757
        Name: connected, dtype: int64

        Fill gaps shorter than 1.1 meters and assign 0 to all columns.

        >>> nw = nw.close_network_holes(max_distance=1.1, fillna=0)
        >>> nw = nw.get_largest_component()
        >>> nw.gdf.connected.value_counts()
        1.0    100315
        0.0       180
        Name: connected, dtype: int64

        It's not always wise to fill gaps. In the case of this data, these small gaps
        are intentional. They are road blocks where most cars aren't allowed to pass.
        Fill the holes only if it makes the travel times/routes more realistic.
        """
        self.gdf = close_network_holes_to_deadends(
            self.gdf,
            max_distance,
            hole_col=hole_col,
        )

        if fillna is not None:
            self.gdf.loc[self.gdf[hole_col] == 1] = self.gdf.loc[
                self.gdf[hole_col] == 1
            ].fillna(fillna)

        self._hole_col = hole_col

        return self

    def cut_lines(
        self,
        max_length: int | float,
        adjust_weight_col: str | None = None,
        ignore_index: bool = False,
    ):
        """Cuts lines into pieces no longer than 'max_length'.

        Will give more accurate results in the service_area method, but not in the
        precice_service_area method. Shorter lines also mean more holes might be filled
        in close_network_holes. Other than that, cutting lines has no impact on network
        analysis except making it slower.

        Args:
            max_length: The maximum length of the line segments.
            adjust_weight_col: Name of a numeric column in the GeoDataFrame to
                be adjusted based on the length of the new lines. For example, if the
                column is "minutes", the minute value will be halved if the line is
                halved in length.
            ignore_index: If True, the resulting axis will be labeled 0, 1, …, n - 1.
                Defaults to False.

        Returns:
            Self

        Raises:
            KeyError: If the specified 'adjust_weight_col' column is not present in the
                GeoDataFrame of the network.

        Note:
            This method is time consuming for large networks and low 'max_length'.

        Examples
        --------
        >>> nw = Network(roads)
        >>> nw.gdf.length.describe().round(1)
        count    93395.0
        mean        41.2
        std         78.5
        min          0.2
        25%         14.0
        50%         27.7
        75%         47.5
        max       5213.7
        dtype: float64
        >>> nw = nw.cut_lines(max_length=100)
        >>> nw.gdf.length.describe().round(1)
        count    126304.0
        mean         30.5
        std          30.1
        min           0.0
        25%           5.7
        50%          22.5
        75%          44.7
        max         100.0
        dtype: float64
        """
        if adjust_weight_col:
            if adjust_weight_col not in self.gdf.columns:
                raise KeyError(f"'gdf' has no column {adjust_weight_col}")
            self.gdf["original_length"] = self.gdf.length

        self.gdf = cut_lines(self.gdf, max_length=max_length, ignore_index=ignore_index)

        if adjust_weight_col:
            self.gdf[adjust_weight_col] = self.gdf[adjust_weight_col] * (
                self.gdf.length / self.gdf[adjust_weight_col]
            )

        return self

    def _make_node_ids(self) -> None:
        """Gives the lines node ids and return lines (edges) and nodes.

        Takes the first and last point of each line and creates a GeoDataFrame of
        nodes (points) with a column 'node_id'. The node ids are then assigned to the
        input GeoDataFrame of lines as the columns 'source' and 'target'.

        Note:
            The lines must be singlepart linestrings.
        """
        self.gdf, self._nodes = make_node_ids(self.gdf)

    @staticmethod
    def _prepare_network(gdf: GeoDataFrame) -> GeoDataFrame:
        """Make sure there are only singlepart LineStrings in the network.

        This is needed when making node-ids based on the lines' endpoints, because
        MultiLineStrings have more than two endpoints, and LinearRings have zero.
        Rename geometry column to 'geometry',

        Args:
            gdf: GeoDataFrame with (multi)line geometries. MultiLineStrings will be
                merged, then exploded if a merge was not possible.

        Returns:
            A GeoDataFrame of line geometries.

        Raises:
            ZeroLinesError: If the GeoDataFrame has 0 rows.
        """
        gdf["idx_orig"] = gdf.index

        if gdf._geometry_column_name != "geometry":
            gdf = gdf.rename_geometry("geometry")

        gdf = clean_geoms(gdf)
        gdf = to_single_geom_type(gdf, geom_type="lines")

        if not len(gdf):
            raise ZeroLinesError

        gdf.geometry = line_merge(gdf.geometry)

        rows_now = len(gdf)
        gdf = gdf.loc[gdf.geom_type != "LinearRing"]

        if diff := rows_now - len(gdf):
            if diff == 1:
                print(f"{diff} LinearRing was removed from the network.")
            else:
                print(f"{diff} LinearRings were removed from the network.")

        rows_now = len(gdf)
        gdf = gdf.explode(ignore_index=True)

        if diff := len(gdf) - rows_now:
            if diff == 1:
                print(
                    "1 multi-geometry was split into single part geometries. "
                    "Minute column(s) will be wrong for these rows."
                )
            else:
                print(
                    f"{diff} multi-geometries were split into single part geometries. "
                    "Minute column(s) will be wrong for these rows."
                )

        gdf["meters"] = gdf.length

        return gdf

    def _check_percent_bidirectional(self) -> int:
        """Road data often have to be duplicated and flipped to make it directed.

        Here we check how.
        """
        self.gdf["meters"] = self.gdf["meters"].astype(str)
        no_dups = DataFrame(
            np.sort(self.gdf[["source", "target", "meters"]].values, axis=1),
            columns=[["source", "target", "meters"]],
        ).drop_duplicates()
        self.gdf["meters"] = self.gdf.length

        percent_bidirectional = len(self.gdf) / len(no_dups) * 100 - 100

        return int(round(percent_bidirectional, 0))

    def _nodes_are_up_to_date(self) -> bool:
        """Check if nodes need to be updated.

        Returns False if there are any source or target values not in the node-ids,
        or any superfluous node-ids (meaning rows have been removed from the lines
        gdf).
        """
        new_or_missing = (~self.gdf.source.isin(self._nodes.node_id)) | (
            ~self.gdf.target.isin(self._nodes.node_id)
        )

        if any(new_or_missing):
            return False

        removed = ~(
            (self._nodes.node_id.isin(self.gdf.source))
            | (self._nodes.node_id.isin(self.gdf.target))
        )

        if any(removed):
            return False

        return True

    def get_edges(self) -> list[tuple[str, str]]:
        return [
            (str(source), str(target))
            for source, target in zip(
                self.gdf["source"], self.gdf["target"], strict=True
            )
        ]

    @staticmethod
    def _create_edge_ids(
        edges: list[tuple[str, str]], weights: list[float]
    ) -> list[str]:
        """Edge identifiers represented with source and target ids and the weight."""
        return [f"{s}_{t}_{w}" for (s, t), w in zip(edges, weights, strict=True)]

    def _update_nodes_if(self):
        if not self._nodes_are_up_to_date():
            self._make_node_ids()

    @property
    def nodes(self):
        """GeoDataFrame with the network nodes (line endpoints).

        Upon instantiation of the class, a GeoDataFrame of points is created from the
        unique endpoints of the lines. The node ids are then used to make the 'source'
        and 'target' columns of the line gdf. The nodes are remade every time the
        geometries of the line gdf changes, since the 'source' and 'target' columns
        must be up to date with the actual line geometries when making the graph.
        """
        return self._nodes

    @property
    def percent_bidirectional(self):
        """The percentage of lines that appear in both directions."""
        return self._percent_bidirectional

    def copy(self):
        """Returns a shallow copy of the class instance."""
        return copy(self)

    def deepcopy(self):
        """Returns a deep copy of the class instance."""
        return deepcopy(self)

    def __repr__(self) -> str:
        """The print representation."""
        cl = self.__class__.__name__
        km = int(sum(self.gdf.length) / 1000)
        return f"{cl}({km} km, undirected)"

    def __iter__(self):
        """So the attributes can be iterated through."""
        return iter(self.__dict__.items())

    def __len__(self):
        return len(self.gdf)



# flake8: noqa: F401
from .geopandas_tools.buffer_dissolve_explode import (
    buff,
    buffdiss,
    buffdissexp,
    buffexp,
    dissexp,
)
from .geopandas_tools.general import (
    clean_clip,
    clean_geoms,
    coordinate_array,
    points_in_bounds,
    random_points,
    random_points_in_polygons,
    to_gdf,
    to_lines,
    to_multipoint,
)
from .geopandas_tools.geometry_types import (
    get_geom_type,
    is_single_geom_type,
    to_single_geom_type,
)
from .geopandas_tools.line_operations import (
    close_network_holes,
    close_network_holes_to_deadends,
    cut_lines,
    cut_lines_once,
    get_component_size,
    get_largest_component,
    make_edge_coords_cols,
    make_edge_wkt_cols,
    make_node_ids,
    split_lines_by_nearest_point,
)
from .geopandas_tools.neighbors import (
    get_all_distances,
    get_k_nearest_neighbors,
    get_neighbor_indices,
)
from .geopandas_tools.overlay import clean_shapely_overlay, overlay, overlay_update
from .geopandas_tools.point_operations import snap_all, snap_within_distance
from .geopandas_tools.polygon_operations import close_all_holes, close_small_holes
from .maps.legend import Legend
from .maps.maps import clipmap, explore, qtm, samplemap
from .maps.thematicmap import ThematicMap
from .networkanalysis.directednetwork import DirectedNetwork
from .networkanalysis.network import Network
from .networkanalysis.networkanalysis import NetworkAnalysis
from .networkanalysis.networkanalysisrules import NetworkAnalysisRules
from .read_parquet import read_parquet_url


try:
    from .dapla import exists, read_geopandas, write_geopandas
except ImportError:
    pass



from geopandas import GeoDataFrame
from pandas import DataFrame
from shapely import force_2d

from ..geopandas_tools.construction import gdf_concat
from .directednetwork import DirectedNetwork
from .network import Network


class NetworkDaplaCar(DirectedNetwork):
    def __init__(
        self,
        gdf: GeoDataFrame | str = "daplasti_nyeste",
        turn_restrictions: bool = True,
        fill_holes: bool = False,
        kommuner: str | tuple[str] | None = None,
        **kwargs,
    ):
        if isinstance(gdf, str):
            gdf = read_network_from_dapla(gdf, kommuner)

        if turn_restrictions:
            self.gdf = self.gdf.loc[gdf.turn_restriction == 0]

        if not fill_holes:
            self.gdf = self.gdf.loc[gdf.hole == 0]

        super().__init__(gdf, **kwargs)


class NetworkDaplaBike(DirectedNetwork):
    def __init__(
        self,
        gdf: GeoDataFrame | str = "daplasti_nyeste",
        speed: int | None = 20,
        fill_holes: bool = False,
        kommuner: str | list | tuple | None = None,
        **kwargs,
    ):
        self.speed = speed

        if isinstance(gdf, str):
            gdf = read_network_from_dapla(gdf, kommuner)

        if not fill_holes:
            gdf = gdf.loc[gdf.hole == 0]

        if self.weight == "minutes":
            self.network["minutes"] = self.network.length / speed * 16.6666667

        super().__init__(gdf, **kwargs)


class NetworkDaplaFoot(Network):
    def __init__(
        self,
        gdf: GeoDataFrame | str = "daplasti_nyeste",
        speed: int | None = 5,
        fill_holes: bool = False,
        kommuner: str | list | tuple | None = None,
        **kwargs,
    ):
        self.speed = speed

        if isinstance(gdf, str):
            gdf = read_network_from_dapla(gdf, kommuner)

        if not fill_holes:
            gdf = gdf.loc[gdf.hole == 0]

        if self.weight == "minutes":
            self.network["minutes"] = self.network.length / speed * 16.6666667

        super().__init__(gdf, **kwargs)


def prepare_network_norway(
    roads: GeoDataFrame,
    out_path: str,
    source_col="fromnode",
    target_col="tonode",
    direction_col="oneway",
    minute_cols=("drivetime_fw", "drivetime_bw"),
    muni_col="municipality",
    turn_restrictions: DataFrame | None = None,
):
    roads.columns = [col.lower() for col in roads.columns]

    roads = (
        roads.to_crs(25833)
        .loc[:, [source_col, target_col, direction_col, "geometry"] + list(minute_cols)]
        .rename(
            columns={
                muni_col: "KOMMUNENR",
            }
        )
    )

    roads["geometry"] = force_2d(roads.geometry)

    nw = DirectedNetwork(roads)

    nw = nw.make_directed_network_norway()

    if turn_restrictions:
        nw.network = find_turn_restrictions(nw.network, turn_restrictions)

    nw = nw.close_network_holes(max_distance=1.1, hole_col="hole")

    nw = nw.get_largest_component()

    write_geopandas(nw.gdf, out_path)

    return nw


def read_network_from_dapla(roads, kommuner):
    if kommuner:
        if isinstance(kommuner, str):
            kommuner = [kommuner]
        filters = [("KOMMUNENR", "in", kommuner)]
    else:
        filters = None

    return read_geopandas(roads, filters=filters)


def find_turn_restrictions(roads, turn_restrictions):
    """Not sure which version is correct because of errors in the source data."""

    # FID starter på  1
    roads["fid"] = roads["idx"] + 1

    turn_restrictions.columns = [col.lower() for col in turn_restrictions.columns]

    for col in turn_restrictions.columns:
        turn_restrictions[col] = turn_restrictions[col].astype(str)

    roads["linkid"] = roads["linkid"].astype(str)
    roads_with_restriction = roads.merge(
        turn_restrictions, left_on="linkid", right_on="fromlinkid", how="inner"
    )
    #    roads_with_restriction = roads.merge(turn_restrictions, left_on = ["source", "target"], right_on = ["fromfromnode", "fromtonode"], how = "inner")
    #    roads_with_restriction2 = roads.merge(turn_restrictions, left_on = "source", "target", "linkid"], right_on = ["fromfromnode", "fromtonode", "fromlinkid"], how = "inner")

    # gjør lenkene med restrictions til første del av nye double_roads som skal lages
    roads_with_restriction = roads_with_restriction.rename(
        columns={
            "target": "middlenode",
            "minutes": "minutes1",
            "meter": "meter1",
            "geometry": "geom1",
        }
    ).loc[:, ["source", "source_wkt", "middlenode", "minutes1", "meter1", "geom1"]]
    # klargjør tabell som skal bli andre del av dobbellenkene
    restrictions = roads.rename(
        columns={
            "source": "middlenode",
            "minutes": "minutes2",
            "meter": "meter2",
            "geometry": "geom2",
        }
    ).loc[:, ["middlenode", "target", "target_wkt", "minutes2", "meter2", "geom2"]]

    # koble basert på den nye kolonnen 'middlenode', som blir midterste node i
    # dobbellenkene
    fra_noder_med_restriction = roads_with_restriction.merge(
        restrictions, on="middlenode", how="inner"
    )

    double_roads = fra_noder_med_restriction[
        ~(
            (
                fra_noder_med_restriction["source"].isin(
                    turn_restrictions["fromfromnode"]
                )
            )
            &
            # (fra_noder_med_restriction["middlenode"].isin(turn_restrictions["fromtonode"])) &
            (fra_noder_med_restriction["target"].isin(turn_restrictions["totonode"]))
        )
    ]

    # smelt lenkeparene sammen
    double_roads["minutes"] = double_roads["minutes1"] + double_roads["minutes2"]
    double_roads["meter"] = double_roads["meter1"] + double_roads["meter2"]
    double_roads["geometry"] = line_merge(union(double_roads.geom1, double_roads.geom2))
    double_roads = gpd.GeoDataFrame(double_roads, geometry="geometry", crs=roads.crs)

    double_roads["turn_restriction"] = 1
    roads.loc[
        (roads["linkid"].isin(turn_restrictions["fromlinkid"])), "turn_restriction"
    ] = 0

    return gdf_concat([roads, double_roads])



"""Interactive and static mapping of multiple GeoDataFrames.

The main function is 'explore', which displays one of more GeoDataFrames together in an
interactive map with layers that can be toggled on and off. The 'samplemap' and
'clipmap' functions do the same, but displays a random and chosen area respectfully.

The 'qtm' function shows a static map of one or more GeoDataFrames.
"""
from geopandas import GeoDataFrame, GeoSeries
from shapely import Geometry

from ..exceptions import NotInJupyterError
from ..geopandas_tools.general import random_points_in_polygons
from ..geopandas_tools.geometry_types import get_geom_type
from ..helpers import make_namedict
from .explore import Explore
from .map import Map
from .thematicmap import ThematicMap


def _check_if_jupyter_is_needed(explore, show_in_browser):
    if explore and not show_in_browser:
        try:
            display
        except NameError as e:
            raise NotInJupyterError(
                "Cannot display interactive map. Try setting "
                "'show_in_browser' to True, or 'explore' to False."
            ) from e


def explore(
    *gdfs: GeoDataFrame,
    column: str | None = None,
    labels: tuple[str] | None = None,
    max_zoom: int = 30,
    show_in_browser: bool = False,
    **kwargs,
) -> None:
    """Interactive map of GeoDataFrames with layers that can be toggled on/off.

    It takes all the given GeoDataFrames and displays them together in an
    interactive map with a common legend. If 'column' is not specified, each
    GeoDataFrame is given a unique color.

    If the column is of type string and only one GeoDataFrame is given, the unique
    values will be split into separate GeoDataFrames so that each value can be toggled
    on/off.

    The coloring can be changed with the 'cmap' parameter. The default colormap is a
    custom, strongly colored palette. If a numerical colum is given, the 'viridis'
    palette is used.

    Note:
        The maximum zoom level only works on the OpenStreetMap background map.

    Args:
        *gdfs: one or more GeoDataFrames.
        column: The column to color the geometries by. Defaults to None, which means
            each GeoDataFrame will get a unique color.
        labels: By default, the GeoDataFrames will be labeled by their object names.
            Alternatively, labels can be specified as a tuple of strings with the same
            length as the number of gdfs.
        max_zoom: The maximum allowed level of zoom. Higher number means more zoom
            allowed. Defaults to 30, which is higher than the geopandas default.
        show_in_browser: If False (default), the maps will be shown in Jupyter.
            If True the maps will be opened in a browser folder.
        **kwargs: Keyword arguments to pass to geopandas.GeoDataFrame.explore, for
            instance 'cmap' to change the colors, 'scheme' to change how the data
            is grouped. This defaults to 'fisherjenkssampled' for numeric data.

    See also
    --------
    samplemap: same functionality, but shows only a random area of a given size.
    clipmap: same functionality, but shows only the areas clipped by a given mask.

    Examples
    --------
    >>> from sgis import read_parquet_url, explore
    >>> roads = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_eidskog_2022.parquet")
    >>> points = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/points_eidskog.parquet")

    Simple explore of two GeoDataFrames.

    >>> explore(roads, points)

    With additional arguments.

    >>> roads["meters"] = roads.length
    >>> points["meters"] = points.length
    >>> explore(roads, points, column="meters", cmap="plasma", max_zoom=60)
    """
    m = Explore(
        *gdfs,
        column=column,
        labels=labels,
        show_in_browser=show_in_browser,
        max_zoom=max_zoom,
        **kwargs,
    )
    m.explore()


def samplemap(
    *gdfs: GeoDataFrame,
    column: str | None = None,
    size: int = 1000,
    sample_from_first: bool = True,
    labels: tuple[str] | None = None,
    max_zoom: int = 30,
    explore: bool = True,
    show_in_browser: bool = False,
    **kwargs,
) -> None:
    """Shows an interactive map of a random area of GeoDataFrames.

    It takes all the GeoDataFrames specified, takes a random sample point from the
    first, and shows all geometries within a given radius of this point. Otherwise
    works like the explore function.

    Note:
        The maximum zoom level only works on the OpenStreetMap background map.

    Args:
        *gdfs: one or more GeoDataFrames.
        column: The column to color the geometries by. Defaults to None, which means
            each GeoDataFrame will get a unique color.
        size: the radius to buffer the sample point by before clipping with the data.
            Defaults to 1000 (meters).
        sample_from_first: If True (default), the sample point is taken form the
            first specified GeoDataFrame. If False, all GeoDataFrames are considered.
        labels: By default, the GeoDataFrames will be labeled by their object names.
            Alternatively, labels can be specified as a tuple of strings the same
            length as the number of gdfs.
        max_zoom: The maximum allowed level of zoom. Higher number means more zoom
            allowed. Defaults to 30, which is higher than the geopandas default.
        explore: If True (default), an interactive map will be displayed. If False,
            or not in Jupyter, a static plot will be shown.
        show_in_browser: If False (default), the maps will be shown in Jupyter.
            If True the maps will be opened in a browser folder.
        **kwargs: Keyword arguments to pass to geopandas.GeoDataFrame.explore, for
            instance 'cmap' to change the colors, 'scheme' to change how the data
            is grouped. This defaults to 'fisherjenkssampled' for numeric data.

    See also
    --------
    explore: Same functionality, but shows the entire area of the geometries.
    clipmap: Same functionality, but shows only the areas clipped by a given mask.

    Examples
    --------
    >>> from sgis import read_parquet_url, samplemap
    >>> roads = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_eidskog_2022.parquet")
    >>> points = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/points_eidskog.parquet")

    With default sample size. To get a new sample area, simply re-run the line.

    >>> samplemap(roads, points)

    Sample area with a radius of 5 kilometers.

    >>> samplemap(roads, points, size=5_000, column="meters")

    """
    _check_if_jupyter_is_needed(explore, show_in_browser)

    if not size and isinstance(gdfs[-1], (float, int)):
        *gdfs, size = gdfs

    if explore:
        m = Explore(
            *gdfs,
            column=column,
            labels=labels,
            show_in_browser=show_in_browser,
            max_zoom=max_zoom,
            **kwargs,
        )
        m.samplemap(size, sample_from_first=sample_from_first)
    else:
        m = Map(
            *gdfs,
            column=column,
            labels=labels,
            **kwargs,
        )

        if sample_from_first:
            sample = m._gdfs[0].sample(1)
        else:
            sample = m._gdf.sample(1)

        # convert lines to polygons
        if get_geom_type(sample) == "line":
            sample["geometry"] = sample.buffer(1)

        if get_geom_type(sample) == "polygon":
            random_point = random_points_in_polygons(sample, 1)

        # if point or mixed geometries
        else:
            random_point = sample.centroid.buffer

        m._gdf = m._gdf.clip(random_point.buffer(size))

        qtm(m._gdf, column=m.column, cmap=m._cmap, k=m.k)


def clipmap(
    *gdfs: GeoDataFrame,
    column: str | None = None,
    mask: GeoDataFrame | GeoSeries | Geometry,
    labels: tuple[str] | None = None,
    explore: bool = True,
    max_zoom: int = 30,
    show_in_browser: bool = False,
    **kwargs,
) -> None:
    """Shows an interactive map of a of GeoDataFrames clipped to the mask extent.

    It takes all the GeoDataFrames specified, clips them to the extent of the mask,
    and displays the resulting geometries in an interactive map with a common legends.
    The layers can be toggled on and off.

    For more info about the labeling and coloring of the map, see the explore function.

    Note:
        The maximum zoom level only works on the OpenStreetMap background map.

    Args:
        *gdfs: one or more GeoDataFrames.
        mask: the geometry to clip the data by.
        column: The column to color the geometries by. Defaults to None, which means
            each GeoDataFrame will get a unique color.
        labels: By default, the GeoDataFrames will be labeled by their object names.
            Alternatively, labels can be specified as a tuple of strings the same
            length as the number of gdfs.
        max_zoom: The maximum allowed level of zoom. Higher number means more zoom
            allowed. Defaults to 30, which is higher than the geopandas default.
        explore: If True (default), an interactive map will be displayed. If False,
            or not in Jupyter, a static plot will be shown.
        show_in_browser: If False (default), the maps will be shown in Jupyter.
            If True the maps will be opened in a browser folder.
        **kwargs: Keyword arguments to pass to geopandas.GeoDataFrame.explore, for
            instance 'cmap' to change the colors, 'scheme' to change how the data
            is grouped. This defaults to 'fisherjenkssampled' for numeric data.

    See also
    --------
    explore: same functionality, but shows the entire area of the geometries.
    samplemap: same functionality, but shows only a random area of a given size.
    """

    _check_if_jupyter_is_needed(explore, show_in_browser)

    gdfs, column = Explore._separate_args(gdfs, column)

    # storing object names in dict here, since the names disappear after clip
    if not labels:
        namedict = make_namedict(gdfs)
        kwargs["namedict"] = namedict

    clipped: tuple[GeoDataFrame] = ()

    if mask is not None:
        for gdf in gdfs:
            clipped_ = gdf.clip(mask)
            clipped = clipped + (clipped_,)

    else:
        for gdf in gdfs[:-1]:
            clipped_ = gdf.clip(gdfs[-1])
            clipped = clipped + (clipped_,)

    if explore:
        m = Explore(
            *clipped,
            column=column,
            labels=labels,
            show_in_browser=show_in_browser,
            max_zoom=max_zoom,
            **kwargs,
        )
        m.explore()
    else:
        m = Map(
            *gdfs,
            column=column,
            labels=labels,
            **kwargs,
        )
        qtm(m._gdf, column=m.column, cmap=m._cmap, k=m.k)


def qtm(
    *gdfs: GeoDataFrame,
    column: str | None = None,
    title: str | None = None,
    black: bool = True,
    size: int = 10,
    legend: bool = True,
    cmap: str | None = None,
    k: int = 5,
    **kwargs,
) -> None:
    """Quick, thematic map of one or more GeoDataFrames.

    Shows one or more GeoDataFrames in the same plot, with a common color scheme if
    column is specified, otherwise with unique colors for each GeoDataFrame.

    The 'qtm' name is taken from the tmap package in R.

    Args:
        *gdfs: One or more GeoDataFrames to plot.
        column: The column to color the map by. Defaults to None, meaning each
            GeoDataFrame is given a unique color.
        title: Text to use as the map's heading.
        black: If True (default), the background color will be black and the title
            white. If False, it will be the opposite. The colormap will also be
            'viridis' when black, and 'RdPu' when white.
        size: Size of the plot. Defaults to 10.
        title_fontsize: Size of the title.
        cmap: Color palette of the map. See:
            https://matplotlib.org/stable/tutorials/colors/colormaps.html
        k: Number of color groups.
        **kwargs: Additional keyword arguments taken by the geopandas plot method.

    See also:
        ThematicMap: Class with more options for customising the plot.
    """

    m = ThematicMap(*gdfs, column=column, size=size, black=black)

    m.title = title

    if k and len(m._unique_values) >= k:
        m.k = k

    if cmap:
        m.cmap = cmap

    if not legend:
        m.legend = None

    m.plot(**kwargs)
